"""
Applies record annotations to overwrite field values.
This does not do any additional transformations on top of the annotations.
"""
import csv
from collections import defaultdict
from augur.errors import AugurError
from augur.io.print import print_err
from augur.utils import first_line


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("apply-record-annotations",
        parents=[parent_subparsers.shared_parser],
        help=first_line(__doc__))

    parser.add_argument("--annotations", metavar="TSV", required=True,
        help="Manually curated annotations TSV file. " +
             "The TSV should not have a header and should have exactly three columns: " +
             "id to match existing metadata, field name, and field value. " +
             "If there are multiple annotations for the same id and field, then the last value is used. " +
             "Lines starting with '#' are treated as comments. " +
             "Any '#' after the field value are treated as comments.")
    parser.add_argument("--id-field", default="accession",
        help="The ID field in the metadata to use to merge with the annotations.")

    return parser


def run(args, records):
    annotations = defaultdict(dict)
    with open(args.annotations, 'r', newline='') as annotations_fh:
        csv_reader = csv.reader(annotations_fh, delimiter='\t')
        for row in csv_reader:
            if not row or row[0].lstrip()[0] == '#':
                    continue
            elif len(row) != 3:
                print_err("WARNING: Could not decode annotation line " + "\t".join(row))
                continue
            id, field, value = row
            annotations[id][field] = value.partition('#')[0].rstrip()

    for record in records:
        record_id = record.get(args.id_field)
        if record_id is None:
            raise AugurError(f"ID field {args.id_field!r} does not exist in record")

        record_annotations = annotations.get(record_id, {})
        for field in list(record_annotations.keys()):
            if field not in record:
                print_err(f"WARNING: Skipping annotation for field {field!r} that does not exist in record")
                del record_annotations[field]

        record.update(record_annotations)

        yield record



"""
Custom helpers for the argparse standard library.
"""
from argparse import Action, ArgumentDefaultsHelpFormatter, ArgumentParser, _ArgumentGroup
from typing import Union
from .types import ValidationMode


# Include this in an argument help string to suppress the automatic appending
# of the default value by argparse.ArgumentDefaultsHelpFormatter.  This works
# because the automatic appending is conditional on the presence of %(default),
# so we include it but then format it as a zero-length string .0s.  🙃
#
# Another solution would be to add an extra attribute to the argument (the
# argparse.Action instance) and then subclass ArgumentDefaultsHelpFormatter to
# condition on that new attribute, but that seems more brittle.
#
# Copied from the Nextstrain CLI repo
# https://github.com/nextstrain/cli/blob/017c53805e8317951327d24c04184615cc400b09/nextstrain/cli/argparse.py#L13-L21
SKIP_AUTO_DEFAULT_IN_HELP = "%(default).0s"


def add_default_command(parser):
    """
    Sets the default command to run when none is provided.
    """
    class default_command():
        def run(args):
            parser.print_help()
            return 2

    parser.set_defaults(__command__ = default_command)


def add_command_subparsers(subparsers, commands, command_attribute='__command__'):
    """
    Add subparsers for each command module.

    Parameters
    ----------
    subparsers: argparse._SubParsersAction
        The special subparsers action object created by the parent parser
        via `parser.add_subparsers()`.

    commands: list[types.ModuleType]
        A list of modules that are commands that require their own subparser.
        Each module is required to have a `register_parser` function to add its own
        subparser and arguments.

    command_attribute: str, optional
        Optional attribute name for the commands. The default is `__command__`,
        which allows top level augur to run commands directly via `args.__command__.run()`.
    """
    for command in commands:
        # Allow each command to register its own subparser
        subparser = command.register_parser(subparsers)

        # Add default attribute for command module
        if command_attribute:
            subparser.set_defaults(**{command_attribute: command})

        # Use the same formatting class for every command for consistency.
        # Set here to avoid repeating it in every command's register_parser().
        subparser.formatter_class = ArgumentDefaultsHelpFormatter

        if not subparser.description and command.__doc__:
            subparser.description = command.__doc__

        # If a command doesn't have its own run() function, then print its help when called.
        if not getattr(command, "run", None):
            add_default_command(subparser)


class HideAsFalseAction(Action):
    """
    Custom argparse Action that stores False for arguments passed as `--hide*`
    and stores True for all other argument patterns.
    """
    def __call__(self, parser, namespace, values, option_string=None):
        setattr(namespace, self.dest, option_string[2:6] != 'hide')


class ExtendOverwriteDefault(Action):
    """
    Similar to the core argparse ``extend`` action, but overwrites the argument
    ``default``, if any, instead of appending to it.

    Thus, the ``default`` value is not included when the option is given and
    may be a non-list value if desired.
    """
    def __call__(self, parser, namespace, value, option_string = None):
        current = getattr(namespace, self.dest, None)

        if current is parser.get_default(self.dest) or current is None:
            current = []

        setattr(namespace, self.dest, [*current, *value])


def add_validation_arguments(parser: Union[ArgumentParser, _ArgumentGroup]):
    """
    Add arguments to configure validation mode of node data JSON files.
    """
    parser.add_argument(
        '--validation-mode',
        dest="validation_mode",
        type=ValidationMode,
        choices=[mode for mode in ValidationMode],
        default=ValidationMode.ERROR,
        help="""
            Control if optional validation checks are performed and what
            happens if they fail.

            'error' and 'warn' modes perform validation and emit messages about
            failed validation checks.  'error' mode causes a non-zero exit
            status if any validation checks failed, while 'warn' does not.

            'skip' mode performs no validation.

            Note that some validation checks are non-optional and as such are
            not affected by this setting.
        """)
    parser.add_argument(
        '--skip-validation',
        dest="validation_mode",
        action="store_const",
        const=ValidationMode.SKIP,
        help="Skip validation of input/output files, equivalent to --validation-mode=skip. Use at your own risk!")



"""
Resource files.
"""
# Originally copied from nextstrain/cli/resources/__init__.py in the Nextstrain
# CLI project¹.
#
# ¹ <https://github.com/nextstrain/cli/blob/a5dda9c0579ece7acbd8e2c32a4bbe95df7c0bce/nextstrain/cli/resources/__init__.py>

# We gate usage of the stdlib implementation on 3.11 because that's the first
# version with a full adapter for making the new files() / Traversable API
# backwards compatible with importers only providing the original path() /
# ResourceReader API.  The PyPI backport, on the other hand, contains the full
# adapter since 5.3.0, which we declare as our minimum version in setup.py, so
# we use that even on 3.9 and 3.10.
#
# We're using the new API at all because the original one is being deprecated
# and we want to avoid warnings both from the stdlib implementation on 3.11 and
# from the PyPI backport implementation on older Python versions.
#   -trs, 13 Sept 2022
import sys

if sys.version_info >= (3, 11):
    from importlib.resources import files as _files, as_file as _as_file
else:
    from importlib_resources import files as _files, as_file as _as_file

from pathlib import Path
from typing import ContextManager


def as_file(path: str) -> ContextManager[Path]:
    return _as_file(_files(__name__) / path)



import Bio.SeqIO

from augur.errors import AugurError
from .file import open_file


def read_sequences(*paths, format="fasta"):
    """Read sequences from one or more paths.

    Automatically infer compression mode (e.g., gzip, etc.) and return a stream
    of sequence records in the requested format (e.g., "fasta", "genbank", etc.).

    Parameters
    ----------
    paths : list of str or `os.PathLike`
        One or more paths to sequence files of any type supported by BioPython.

    format : str
        Format of input sequences matching any of those supported by BioPython
        (e.g., "fasta", "genbank", etc.).

    Yields
    ------
    Bio.SeqRecord.SeqRecord
        Sequence record from the given path(s).

    """
    for path in paths:
        # Open the given path as a handle, inferring the file's compression.
        # This way we can pass a handle to BioPython's SeqIO interface
        # regardless of the compression mode.
        with open_file(path) as handle:
            sequences = Bio.SeqIO.parse(handle, format)

            for sequence in sequences:
                yield sequence


def write_sequences(sequences, path_or_buffer, format="fasta"):
    """Write sequences to a given path in the given format.

    Automatically infer compression mode (e.g., gzip, etc.) based on the path's
    filename extension.

    Parameters
    ----------
    sequences : iterable of Bio.SeqRecord.SeqRecord
        A list-like collection of sequences to write

    path_or_buffer : str or `os.PathLike` or `io.StringIO`
        A path to a file to write the given sequences in the given format.

    format : str
        Format of input sequences matching any of those supported by BioPython
        (e.g., "fasta", "genbank", etc.)

    Returns
    -------
    int :
        Number of sequences written out to the given path.

    """
    with open_file(path_or_buffer, "wt") as handle:
        # Bio.SeqIO supports writing to the same handle multiple times for specific
        # file formats. For the formats we use, this function call should work for
        # both a newly opened file handle or one that is provided by the caller.
        # For more details see:
        # https://github.com/biopython/biopython/blob/25f5152f4aeefe184a323db25694fbfe0593f0e2/Bio/SeqIO/__init__.py#L233-L251
        sequences_written = Bio.SeqIO.write(
            sequences,
            handle,
            format
        )

    return sequences_written


def write_records_to_fasta(records, fasta, seq_id_field='strain', seq_field='sequence'):
    """
    Write sequences from dict *records* to a *fasta* file.
    Yields the records with the *seq_field* dropped so that they can be consumed downstream.

    Parameters
    ----------
    records: iterable of dict
        Iterator that yields dict that contains sequences

    fasta: str
        Path to FASTA file

    seq_id_field: str, optional
        Field name for the sequence identifier

    seq_field: str, optional
        Field name for the genomic sequence

    Yields
    ------
    dict:
        A copy of the record with *seq_field* dropped

    Raises
    ------
    AugurError
        When the sequence id field or sequence field does not exist in a record
    """
    with open_file(fasta, "w") as output_fasta:
        for record in records:
            if seq_id_field not in record:
                raise AugurError(f"Provided sequence identifier field {seq_id_field!r} does not exist.")
            if seq_field not in record:
                raise AugurError(f"Provided sequence field {seq_field!r} does not exist.")

            output_fasta.writelines([
                f">{record[seq_id_field]}\n",
                f"{record[seq_field]}\n"
            ])

            yield {
                key: value
                for key, value in record.items()
                if key != seq_field
            }



from collections import defaultdict
from contextlib import nullcontext
import csv
import itertools
import json
import numpy as np
import os
import pandas as pd
from tempfile import NamedTemporaryFile

from augur.errors import AugurError
from augur.index import (
    index_sequences,
    index_vcf,
    ID_COLUMN as SEQUENCE_INDEX_ID_COLUMN,
    DELIMITER as SEQUENCE_INDEX_DELIMITER,
)
from augur.io.file import PANDAS_READ_CSV_OPTIONS, open_file
from augur.io.metadata import InvalidDelimiter, Metadata, read_metadata
from augur.io.sequences import read_sequences, write_sequences
from augur.io.print import print_err, _n
from augur.io.vcf import is_vcf as filename_is_vcf, write_vcf
from augur.types import EmptyOutputReportingMethod
from . import include_exclude_rules
from .io import cleanup_outputs, get_useful_metadata_columns, read_priority_scores, write_metadata_based_outputs
from .include_exclude_rules import apply_filters, construct_filters
from .subsample import PriorityQueue, TooManyGroupsError, calculate_sequences_per_group, get_probabilistic_group_sizes, create_queues_by_group, get_groups_for_subsampling, get_weighted_group_sizes


def run(args):
    # Determine whether the sequence index exists or whether should be
    # generated. We need to generate an index if the input sequences are in a
    # VCF, if sequence output has been requested (so we can filter strains by
    # sequences that are present), or if any other sequence-based filters have
    # been requested.
    sequence_strains = None
    sequence_index_path = args.sequence_index
    build_sequence_index = False
    is_vcf = filename_is_vcf(args.sequences)

    # Don't build sequence index with --exclude-all since the only way to add
    # strains back in with this flag are the `--include` or `--include-where`
    # options, so we know we don't need a sequence index to apply any additional
    # filters.
    if sequence_index_path is None and args.sequences and not args.exclude_all:
        build_sequence_index = True

    if build_sequence_index:
        # Generate the sequence index on the fly, for backwards compatibility
        # with older workflows that don't generate the index ahead of time.
        # Create a temporary index using a random filename to avoid collisions
        # between multiple filter commands.
        with NamedTemporaryFile(delete=False) as sequence_index_file:
            sequence_index_path = sequence_index_file.name

        print_err(
            "Note: You did not provide a sequence index, so Augur will generate one.",
            "You can generate your own index ahead of time with `augur index` and pass it with `augur filter --sequence-index`."
        )

        if is_vcf:
            index_vcf(args.sequences, sequence_index_path)
        else:
            index_sequences(args.sequences, sequence_index_path)

    # Load the sequence index, if a path exists.
    sequence_index = None
    if sequence_index_path:
        sequence_index = pd.read_csv(
            sequence_index_path,
            sep=SEQUENCE_INDEX_DELIMITER,
            index_col=SEQUENCE_INDEX_ID_COLUMN,
            dtype={SEQUENCE_INDEX_ID_COLUMN: "string"},
            **PANDAS_READ_CSV_OPTIONS,
        )

        # Remove temporary index file, if it exists.
        if build_sequence_index:
            os.unlink(sequence_index_path)

        sequence_strains = set(sequence_index.index.values)

    #####################################
    #Filtering steps
    #####################################

    # Setup filters.
    exclude_by, include_by = construct_filters(
        args,
        sequence_index,
    )

    # Setup grouping. We handle the following major use cases:
    #
    # 1. group by and sequences per group defined -> use the given values by the
    # user to identify the highest priority records from each group in a single
    # pass through the metadata.
    #
    # 2. group by and maximum sequences defined -> use the first pass through
    # the metadata to count the number of records in each group, calculate the
    # sequences per group that satisfies the requested maximum, and use a second
    # pass through the metadata to select that many sequences per group.
    #
    # 3. group by not defined but maximum sequences defined -> use a "dummy"
    # group such that we select at most the requested maximum number of
    # sequences in a single pass through the metadata.
    #
    # Each case relies on a priority queue to track the highest priority records
    # per group. In the best case, we can track these records in a single pass
    # through the metadata. In the worst case, we don't know how many sequences
    # per group to use, so we need to calculate this number after the first pass
    # and use a second pass to add records to the queue.
    group_by = args.group_by
    sequences_per_group = args.sequences_per_group
    records_per_group = None

    if group_by and args.subsample_max_sequences:
        # In this case, we need two passes through the metadata with the first
        # pass used to count the number of records per group.
        records_per_group = defaultdict(int)
    elif not group_by and args.subsample_max_sequences:
        group_by = ("_dummy",)
        sequences_per_group = args.subsample_max_sequences

    # If we are grouping data, use queues to store the highest priority strains
    # for each group. When no priorities are provided, they will be randomly
    # generated.
    queues_by_group = None
    if group_by:
        # Use user-defined priorities, if possible. Otherwise, setup a
        # corresponding dictionary that returns a random float for each strain.
        if args.priority:
            priorities = read_priority_scores(args.priority)
        else:
            random_generator = np.random.default_rng(args.subsample_seed)
            priorities = defaultdict(random_generator.random)

    # Setup logging.
    output_log_context_manager = open_file(args.output_log, "w", newline='')
    output_log_writer = None
    if args.output_log:
        # Log the names of strains that were filtered or force-included, so we
        # can properly account for each strain (e.g., including those that were
        # initially filtered for one reason and then included again for another
        # reason).
        output_log = output_log_context_manager.__enter__()
        output_log_header = ("strain", "filter", "kwargs")
        output_log_writer = csv.DictWriter(
            output_log,
            fieldnames=output_log_header,
            delimiter="\t",
            lineterminator="\n",
        )
        output_log_writer.writeheader()

    # Load metadata. Metadata are the source of truth for which sequences we
    # want to keep in filtered output.
    metadata_strains = set()
    valid_strains = set() # TODO: rename this more clearly
    all_sequences_to_include = set()
    filter_counts = defaultdict(int)

    try:
        metadata_object = Metadata(args.metadata, args.metadata_delimiters, args.metadata_id_columns)
    except InvalidDelimiter:
        raise AugurError(
            f"Could not determine the delimiter of {args.metadata!r}. "
            f"Valid delimiters are: {args.metadata_delimiters!r}. "
            "This can be changed with --metadata-delimiters."
        )
    useful_metadata_columns = get_useful_metadata_columns(args, metadata_object.id_column, metadata_object.columns)

    metadata_reader = read_metadata(
        args.metadata,
        delimiters=[metadata_object.delimiter],
        columns=useful_metadata_columns,
        id_columns=[metadata_object.id_column],
        chunk_size=args.metadata_chunk_size,
        dtype="string",
    )
    for metadata in metadata_reader:
        duplicate_strains = (
            set(metadata.index[metadata.index.duplicated()]) |
            (set(metadata.index) & metadata_strains)
        )
        if len(duplicate_strains) > 0:
            cleanup_outputs(args)
            raise AugurError(f"The following strains are duplicated in '{args.metadata}':\n" + "\n".join(sorted(duplicate_strains)))

        # Maintain list of all strains seen.
        metadata_strains.update(set(metadata.index.values))

        # Filter metadata.
        seq_keep, sequences_to_filter, sequences_to_include = apply_filters(
            metadata,
            exclude_by,
            include_by,
        )
        valid_strains.update(seq_keep)

        # Track distinct strains to include, so we can write their
        # corresponding metadata, strains, or sequences later, as needed.
        distinct_force_included_strains = {
            record["strain"]
            for record in sequences_to_include
        }
        all_sequences_to_include.update(distinct_force_included_strains)

        # Track reasons for filtered or force-included strains, so we can
        # report total numbers filtered and included at the end. Optionally,
        # write out these reasons to a log file.
        for filtered_strain in itertools.chain(sequences_to_filter, sequences_to_include):
            filter_counts[(filtered_strain["filter"], filtered_strain["kwargs"])] += 1

            # Log the names of strains that were filtered or force-included,
            # so we can properly account for each strain (e.g., including
            # those that were initially filtered for one reason and then
            # included again for another reason).
            if args.output_log:
                output_log_writer.writerow(filtered_strain)

        if group_by:
            # Prevent force-included sequences from being included again during
            # subsampling.
            seq_keep = seq_keep - distinct_force_included_strains

            # If grouping, track the highest priority metadata records or
            # count the number of records per group. First, we need to get
            # the groups for the given records.
            group_by_strain = get_groups_for_subsampling(
                seq_keep,
                metadata,
                group_by,
            )

            if args.subsample_max_sequences and records_per_group is not None:
                # Count the number of records per group. We will use this
                # information to calculate the number of sequences per group
                # for the given maximum number of requested sequences.
                for group in group_by_strain.values():
                    records_per_group[group] += 1
            else:
                # Track the highest priority records, when we already
                # know the number of sequences allowed per group.
                if queues_by_group is None:
                    queues_by_group = {}

                for strain in sorted(group_by_strain.keys()):
                    # During this first pass, we do not know all possible
                    # groups will be, so we need to build each group's queue
                    # as we first encounter the group.
                    group = group_by_strain[strain]
                    if group not in queues_by_group:
                        queues_by_group[group] = PriorityQueue(
                            max_size=sequences_per_group,
                        )

                    queues_by_group[group].add(
                        metadata.loc[strain],
                        priorities[strain],
                    )

    # In the worst case, we need to calculate sequences per group from the
    # requested maximum number of sequences and the number of sequences per
    # group. Then, we need to make a second pass through the metadata to find
    # the requested number of records.
    if args.subsample_max_sequences and records_per_group is not None:
        if queues_by_group is None:
            # We know all of the possible groups now from the first pass through
            # the metadata, so we can create queues for all groups at once.
            if args.group_by_weights:
                print_err(f"Sampling with weights defined by {args.group_by_weights}.")
                group_sizes = get_weighted_group_sizes(
                    records_per_group,
                    group_by,
                    args.group_by_weights,
                    args.subsample_max_sequences,
                    args.output_group_by_sizes,
                    args.subsample_seed,
                )
            else:
                # Calculate sequences per group. If there are more groups than maximum
                # sequences requested, sequences per group will be a floating point
                # value and subsampling will be probabilistic.
                try:
                    sequences_per_group, probabilistic_used = calculate_sequences_per_group(
                        args.subsample_max_sequences,
                        records_per_group.values(),
                        args.probabilistic_sampling,
                    )
                except TooManyGroupsError as error:
                    raise AugurError(error)

                if (probabilistic_used):
                    print_err(f"Sampling probabilistically at {sequences_per_group:0.4f} sequences per group, meaning it is possible to have more than the requested maximum of {args.subsample_max_sequences} sequences after filtering.")
                    group_sizes = get_probabilistic_group_sizes(
                        records_per_group.keys(),
                        sequences_per_group,
                        random_seed=args.subsample_seed,
                    )
                else:
                    print_err(f"Sampling at {sequences_per_group} per group.")
                    assert type(sequences_per_group) is int
                    group_sizes = {group: sequences_per_group for group in records_per_group.keys()}
            queues_by_group = create_queues_by_group(group_sizes)

        # Make a second pass through the metadata, only considering records that
        # have passed filters.
        metadata_reader = read_metadata(
            args.metadata,
            delimiters=args.metadata_delimiters,
            columns=useful_metadata_columns,
            id_columns=args.metadata_id_columns,
            chunk_size=args.metadata_chunk_size,
            dtype="string",
        )
        for metadata in metadata_reader:
            # Recalculate groups for subsampling as we loop through the
            # metadata a second time. TODO: We could store these in memory
            # during the first pass, but we want to minimize overall memory
            # usage at the moment.
            seq_keep = set(metadata.index.values) & valid_strains

            # Prevent force-included strains from being considered in this
            # second pass, as in the first pass.
            seq_keep = seq_keep - all_sequences_to_include

            group_by_strain = get_groups_for_subsampling(
                seq_keep,
                metadata,
                group_by,
            )

            for strain in sorted(group_by_strain.keys()):
                group = group_by_strain[strain]
                queues_by_group[group].add(
                    metadata.loc[strain],
                    priorities[strain],
                )

    # If we have any records in queues, we have grouped results and need to
    # stream the highest priority records to the requested outputs.
    num_excluded_subsamp = 0
    if queues_by_group:
        # Populate the set of strains to keep from the records in queues.
        subsampled_strains = set()
        for group, queue in queues_by_group.items():
            records = []
            for record in queue.get_items():
                # Each record is a pandas.Series instance. Track the name of the
                # record, so we can output its sequences later.
                subsampled_strains.add(record.name)

                # Construct a data frame of records to simplify metadata output.
                records.append(record)

        # Count and optionally log strains that were not included due to
        # subsampling.
        strains_filtered_by_subsampling = valid_strains - subsampled_strains
        num_excluded_subsamp = len(strains_filtered_by_subsampling)
        if output_log_writer:
            for strain in strains_filtered_by_subsampling:
                output_log_writer.writerow({
                    "strain": strain,
                    "filter": "subsampling",
                    "kwargs": "",
                })

        valid_strains = subsampled_strains

    # Force inclusion of specific strains after filtering and subsampling.
    valid_strains = valid_strains | all_sequences_to_include

    # Write output starting with sequences, if they've been requested. It is
    # possible for the input sequences and sequence index to be out of sync
    # (e.g., the index is a superset of the given sequences input), so we need
    # to update the set of strains to keep based on which strains are actually
    # available.
    if is_vcf:
        if args.output:
            # Get the samples to be deleted, not to keep, for VCF
            dropped_samps = list(sequence_strains - valid_strains)
            write_vcf(args.sequences, args.output, dropped_samps)
    elif args.sequences:
        # If the user requested sequence output, stream to disk all sequences
        # that passed all filters to avoid reading sequences into memory first.
        # Even if we aren't emitting sequences, we check for duplicates and
        # track the observed strain names in the sequence file as part of the
        # single pass to allow comparison with the provided sequence index.
        observed_sequence_strains = set()
        duplicates = set()
        with open_file(args.output, "wt") if args.output else nullcontext() as output_handle:
            for sequence in read_sequences(args.sequences):
                if sequence.id in observed_sequence_strains:
                    duplicates.add(sequence.id)

                observed_sequence_strains.add(sequence.id)

                if args.output:
                    if sequence.id in valid_strains:
                        write_sequences(sequence, output_handle, 'fasta')

        if duplicates:
            cleanup_outputs(args)
            raise AugurError(f"The following strains are duplicated in '{args.sequences}':\n" + "\n".join(sorted(duplicates)))

        if sequence_strains != observed_sequence_strains:
            # Warn the user if the expected strains from the sequence index are
            # not a superset of the observed strains.
            if sequence_strains is not None and observed_sequence_strains > sequence_strains:
                print_err(
                    "WARNING: The sequence index is out of sync with the provided sequences.",
                    "Metadata and strain output may not match sequence output."
                )

            # Update the set of available sequence strains.
            sequence_strains = observed_sequence_strains

    if args.output_metadata or args.output_strains:
        write_metadata_based_outputs(args.metadata, args.metadata_delimiters,
                                     args.metadata_id_columns, args.output_metadata,
                                     args.output_strains, valid_strains)

    # Calculate the number of strains that don't exist in either metadata or
    # sequences.
    num_excluded_by_lack_of_metadata = 0
    if sequence_strains:
        num_excluded_by_lack_of_metadata = len(sequence_strains - metadata_strains)


    # Calculate the number of strains passed and filtered.
    total_strains_passed = len(valid_strains)
    total_strains_filtered = len(metadata_strains) + num_excluded_by_lack_of_metadata - total_strains_passed

    print_err(f"{total_strains_filtered} {_n('strain was', 'strains were', total_strains_filtered)} dropped during filtering")

    if num_excluded_by_lack_of_metadata:
        print_err(f"\t{num_excluded_by_lack_of_metadata} had no metadata")

    report_template_by_filter_name = {
        include_exclude_rules.filter_by_sequence_index.__name__: "{count} had no sequence data",
        include_exclude_rules.filter_by_exclude_all.__name__: "{count} {were} dropped by `--exclude-all`",
        include_exclude_rules.filter_by_exclude.__name__: "{count} {were} dropped because {they} {were} in {exclude_file}",
        include_exclude_rules.filter_by_exclude_where.__name__: "{count} {were} dropped because of '{exclude_where}'",
        include_exclude_rules.filter_by_query.__name__: "{count} {were} filtered out by the query: \"{query}\"",
        include_exclude_rules.filter_by_ambiguous_date.__name__: "{count} {were} dropped because of their ambiguous date in {ambiguity}",
        include_exclude_rules.filter_by_min_date.__name__: "{count} {were} dropped because {they} {were} earlier than {min_date} or missing a date",
        include_exclude_rules.filter_by_max_date.__name__: "{count} {were} dropped because {they} {were} later than {max_date} or missing a date",
        include_exclude_rules.filter_by_min_length.__name__: "{count} {were} dropped because {they} {were} shorter than the minimum length of {min_length}bp when only counting standard nucleotide characters A, C, G, or T (case-insensitive)",
        include_exclude_rules.filter_by_max_length.__name__: "{count} {were} dropped because {they} {were} longer than the maximum length of {max_length}bp when only counting standard nucleotide characters A, C, G, or T (case-insensitive)",
        include_exclude_rules.filter_by_non_nucleotide.__name__: "{count} {were} dropped because {they} had non-nucleotide characters",
        include_exclude_rules.skip_group_by_with_ambiguous_year.__name__: "{count} {were} dropped during grouping due to ambiguous year information",
        include_exclude_rules.skip_group_by_with_ambiguous_month.__name__: "{count} {were} dropped during grouping due to ambiguous month information",
        include_exclude_rules.skip_group_by_with_ambiguous_day.__name__: "{count} {were} dropped during grouping due to ambiguous day information",
        include_exclude_rules.force_include_strains.__name__: "{count} {were} added back because {they} {were} in {include_file}",
        include_exclude_rules.force_include_where.__name__: "{count} {were} added back because of '{include_where}'",
    }
    for (filter_name, filter_kwargs), count in filter_counts.items():
        if filter_kwargs:
            parameters = dict(json.loads(filter_kwargs))
        else:
            parameters = {}

        parameters["count"] = count
        parameters["were"] = _n("was", "were", count)
        parameters["they"] = _n("it", "they", count)
        print_err("\t" + report_template_by_filter_name[filter_name].format(**parameters))

    if (group_by and args.sequences_per_group) or args.subsample_max_sequences:
        seed_txt = ", using seed {}".format(args.subsample_seed) if args.subsample_seed else ""
        print_err(f"\t{num_excluded_subsamp} {_n('was', 'were', num_excluded_subsamp)} dropped because of subsampling criteria{seed_txt}")

    if total_strains_passed == 0:
        empty_results_message = "All samples have been dropped! Check filter rules and metadata file format."
        if args.empty_output_reporting is EmptyOutputReportingMethod.ERROR:
            raise AugurError(empty_results_message)
        elif args.empty_output_reporting is EmptyOutputReportingMethod.WARN:
            print_err(f"WARNING: {empty_results_message}")
        elif args.empty_output_reporting is EmptyOutputReportingMethod.SILENT:
            pass
        else:
            raise ValueError(f"Encountered unhandled --empty-output-reporting method {args.empty_output_reporting!r}")

    print_err(f"{total_strains_passed} {_n('strain', 'strains', total_strains_passed)} passed all filters")



"""
Abbreviates a full list of authors to be '<first author> et al.' of the NDJSON
record from stdin and outputs modified records to stdout.

Note: This is a "best effort" approach and can potentially mangle the author name.
"""

import argparse
import re
from typing import Generator, List
from augur.io.print import print_err
from augur.utils import first_line


def parse_authors(
    record: dict,
    authors_field: str,
    default_value: str,
    index: int,
    abbr_authors_field: str = None,
) -> dict:
    # Strip and normalize whitespace
    new_authors = re.sub(r"\s+", " ", record[authors_field])

    if new_authors == "":
        new_authors = default_value
    else:
        # Split authors list on comma/semicolon
        # OR "and"/"&" with at least one space before and after
        new_authors = re.split(r"(?:\s*[,，;；]\s*|\s+(?:and|&)\s+)", new_authors)[0]

        # if it does not already end with " et al.", add it
        if not new_authors.strip(". ").endswith(" et al"):
            new_authors += " et al."

    if abbr_authors_field:
        if record.get(abbr_authors_field):
            print_err(
                f"WARNING: the {abbr_authors_field!r} field already exists",
                f"in record {index} and will be overwritten!",
            )

        record[abbr_authors_field] = new_authors
    else:
        record[authors_field] = new_authors

    return record


def register_parser(
    parent_subparsers: argparse._SubParsersAction,
) -> argparse._SubParsersAction:
    parser = parent_subparsers.add_parser(
        "abbreviate-authors",
        parents=[parent_subparsers.shared_parser],  # type: ignore
        help=first_line(__doc__),
    )

    parser.add_argument(
        "--authors-field",
        default="authors",
        help="The field containing list of authors.",
    )
    parser.add_argument(
        "--default-value",
        default="?",
        help="Default value to use if authors list is empty.",
    )
    parser.add_argument(
        "--abbr-authors-field",
        help="The field for the generated abbreviated authors. "
        + "If not provided, the original authors field will be modified.",
    )

    return parser


def run(args: argparse.Namespace, records: List[dict]) -> Generator[dict, None, None]:
    for index, record in enumerate(records):
        parse_authors(
            record,
            args.authors_field,
            args.default_value,
            index,
            args.abbr_authors_field,
        )

        yield record



from collections import defaultdict
import heapq
import itertools
import uuid
import numpy as np
import pandas as pd
from textwrap import dedent
from typing import Collection, Dict, Iterable, List, Optional, Set, Tuple, Union

from augur.dates import get_year_month, get_year_week
from augur.errors import AugurError
from augur.io.metadata import METADATA_DATE_COLUMN
from augur.io.print import print_err, _n
from . import constants
from .weights_file import WEIGHTS_COLUMN, COLUMN_VALUE_FOR_DEFAULT_WEIGHT, get_default_weight, get_weighted_columns, read_weights_file

Group = Tuple[str, ...]
"""Combination of grouping column values in tuple form."""


def get_groups_for_subsampling(strains, metadata, group_by=None):
    """Return a list of groups for each given strain based on the corresponding
    metadata and group by column.

    Parameters
    ----------
    strains : list
        A list of strains to get groups for.
    metadata : pandas.DataFrame
        Metadata to inspect for the given strains.
    group_by : list
        A list of metadata (or generated) columns to group records by.

    Returns
    -------
    dict :
        A mapping of strain names to tuples corresponding to the values of the strain's group.

    Examples
    --------
    >>> strains = ["strain1", "strain2"]
    >>> metadata = pd.DataFrame([{"strain": "strain1", "date": "2020-01-01", "region": "Africa"}, {"strain": "strain2", "date": "2020-02-01", "region": "Europe"}]).set_index("strain")
    >>> group_by = ["region"]
    >>> group_by_strain = get_groups_for_subsampling(strains, metadata, group_by)
    >>> group_by_strain
    {'strain1': ('Africa',), 'strain2': ('Europe',)}

    If we group by year or month, these groups are generated from the date
    string.

    >>> group_by = ["year", "month"]
    >>> group_by_strain = get_groups_for_subsampling(strains, metadata, group_by)
    >>> group_by_strain
    {'strain1': (2020, '2020-01'), 'strain2': (2020, '2020-02')}

    If we omit the grouping columns, the result will group by a dummy column.

    >>> group_by_strain = get_groups_for_subsampling(strains, metadata)
    >>> group_by_strain
    {'strain1': ('_dummy',), 'strain2': ('_dummy',)}

    If we try to group by columns that don't exist, we get an error.

    >>> group_by = ["missing_column"]
    >>> get_groups_for_subsampling(strains, metadata, group_by)
    Traceback (most recent call last):
      ...
    augur.errors.AugurError: The specified group-by categories (['missing_column']) were not found.

    If we try to group by some columns that exist and some that don't, we allow
    grouping to continue and print a warning message to stderr.

    >>> group_by = ["year", "month", "missing_column"]
    >>> group_by_strain = get_groups_for_subsampling(strains, metadata, group_by)
    >>> group_by_strain
    {'strain1': (2020, '2020-01', 'unknown'), 'strain2': (2020, '2020-02', 'unknown')}

    We can group metadata without any non-ID columns.

    >>> metadata = pd.DataFrame([{"strain": "strain1"}, {"strain": "strain2"}]).set_index("strain")
    >>> get_groups_for_subsampling(strains, metadata, group_by=('_dummy',))
    {'strain1': ('_dummy',), 'strain2': ('_dummy',)}
    """
    metadata = metadata.loc[list(strains)]
    group_by_strain = {}

    if len(metadata) == 0:
        return group_by_strain

    if not group_by or group_by == ('_dummy',):
        group_by_strain = {strain: ('_dummy',) for strain in strains}
        return group_by_strain

    group_by_set = set(group_by)
    generated_columns_requested = constants.GROUP_BY_GENERATED_COLUMNS & group_by_set

    # If we could not find any requested categories, we cannot complete subsampling.
    if METADATA_DATE_COLUMN not in metadata and group_by_set <= constants.GROUP_BY_GENERATED_COLUMNS:
        raise AugurError(f"The specified group-by categories ({group_by}) were not found. Note that using any of {sorted(constants.GROUP_BY_GENERATED_COLUMNS)} requires a column called {METADATA_DATE_COLUMN!r}.")
    if not group_by_set & (set(metadata.columns) | constants.GROUP_BY_GENERATED_COLUMNS):
        raise AugurError(f"The specified group-by categories ({group_by}) were not found.")

    # Warn/error based on other columns grouped with week.
    if constants.DATE_WEEK_COLUMN in group_by_set:
        if constants.DATE_YEAR_COLUMN in group_by_set:
            print_err(f"WARNING: {constants.DATE_YEAR_COLUMN!r} grouping will be ignored since {constants.DATE_WEEK_COLUMN!r} includes ISO year.")
            group_by.remove(constants.DATE_YEAR_COLUMN)
            group_by_set.remove(constants.DATE_YEAR_COLUMN)
            generated_columns_requested.remove(constants.DATE_YEAR_COLUMN)
        if constants.DATE_MONTH_COLUMN in group_by_set:
            raise AugurError(f"{constants.DATE_MONTH_COLUMN!r} and {constants.DATE_WEEK_COLUMN!r} grouping cannot be used together.")

    if generated_columns_requested:

        if METADATA_DATE_COLUMN not in metadata:
            # Set generated columns to 'unknown'.
            print_err(f"WARNING: A {METADATA_DATE_COLUMN!r} column could not be found to group-by {sorted(generated_columns_requested)}.")
            print_err(f"Filtering by group may behave differently than expected!")
            df_dates = pd.DataFrame({col: 'unknown' for col in constants.GROUP_BY_GENERATED_COLUMNS}, index=metadata.index)
            metadata = pd.concat([metadata, df_dates], axis=1)
        else:
            # Create a DataFrame with year/month/day columns as nullable ints.
            # These columns are prefixed to note temporary usage. They are used
            # to generate other columns, and will be discarded at the end.
            temp_prefix = str(uuid.uuid4())
            temp_date_cols = [f'{temp_prefix}year', f'{temp_prefix}month', f'{temp_prefix}day']
            df_dates = metadata[METADATA_DATE_COLUMN].str.split('-', n=2, expand=True)
            df_dates = df_dates.set_axis(temp_date_cols[:len(df_dates.columns)], axis=1)
            missing_date_cols = set(temp_date_cols) - set(df_dates.columns)
            for col in missing_date_cols:
                df_dates[col] = pd.NA
            for col in temp_date_cols:
                df_dates[col] = pd.to_numeric(df_dates[col], errors='coerce').astype(pd.Int64Dtype())

            # Extend metadata with generated date columns
            # Drop the date column since it should not be used for grouping.
            metadata = pd.concat([metadata.drop(METADATA_DATE_COLUMN, axis=1), df_dates], axis=1)

            # Check again if metadata is empty after dropping ambiguous dates.
            if metadata.empty:
                return group_by_strain

            # Generate columns.
            if constants.DATE_YEAR_COLUMN in generated_columns_requested:
                metadata[constants.DATE_YEAR_COLUMN] = metadata[f'{temp_prefix}year']
            if constants.DATE_MONTH_COLUMN in generated_columns_requested:
                metadata[constants.DATE_MONTH_COLUMN] = metadata.apply(lambda row: get_year_month(
                    row[f'{temp_prefix}year'],
                    row[f'{temp_prefix}month']
                    ), axis=1
                )
            if constants.DATE_WEEK_COLUMN in generated_columns_requested:
                # Note that week = (year, week) from the date.isocalendar().
                # Do not combine the raw year with the ISO week number alone,
                # since raw year ≠ ISO year.
                metadata[constants.DATE_WEEK_COLUMN] = metadata.apply(lambda row: get_year_week(
                    row[f'{temp_prefix}year'],
                    row[f'{temp_prefix}month'],
                    row[f'{temp_prefix}day']
                    ), axis=1
                )

            # Drop the internally used columns.
            for col in temp_date_cols:
                metadata.drop(col, axis=1, inplace=True)

    unknown_groups = group_by_set - set(metadata.columns)
    if unknown_groups:
        print_err(f"WARNING: Some of the specified group-by categories couldn't be found: {', '.join(unknown_groups)}")
        print_err("Filtering by group may behave differently than expected!")
        for group in unknown_groups:
            metadata[group] = 'unknown'

    # Finally, determine groups.
    group_by_strain = dict(zip(metadata.index, metadata[group_by].apply(tuple, axis=1)))
    return group_by_strain


class PriorityQueue:
    """A priority queue implementation that automatically replaces lower priority
    items in the heap with incoming higher priority items.

    Examples
    --------

    Add a single record to a heap with a maximum of 2 records.

    >>> queue = PriorityQueue(max_size=2)
    >>> queue.add({"strain": "strain1"}, 0.5)
    1

    Add another record with a higher priority. The queue should be at its maximum
    size.

    >>> queue.add({"strain": "strain2"}, 1.0)
    2
    >>> queue.heap
    [(0.5, 0, {'strain': 'strain1'}), (1.0, 1, {'strain': 'strain2'})]
    >>> list(queue.get_items())
    [{'strain': 'strain1'}, {'strain': 'strain2'}]

    Add a higher priority record that causes the queue to exceed its maximum
    size. The resulting queue should contain the two highest priority records
    after the lowest priority record is removed.

    >>> queue.add({"strain": "strain3"}, 2.0)
    2
    >>> list(queue.get_items())
    [{'strain': 'strain2'}, {'strain': 'strain3'}]

    Add a record with the same priority as another record, forcing the duplicate
    to be resolved by removing the oldest entry.

    >>> queue.add({"strain": "strain4"}, 1.0)
    2
    >>> list(queue.get_items())
    [{'strain': 'strain4'}, {'strain': 'strain3'}]

    """
    def __init__(self, max_size):
        """Create a fixed size heap (priority queue)

        """
        self.max_size = max_size
        self.heap = []
        self.counter = itertools.count()

    def add(self, item, priority):
        """Add an item to the queue with a given priority.

        If adding the item causes the queue to exceed its maximum size, replace
        the lowest priority item with the given item. The queue stores items
        with an additional heap id value (a count) to resolve ties between items
        with equal priority (favoring the most recently added item).

        """
        heap_id = next(self.counter)

        if len(self.heap) >= self.max_size:
            heapq.heappushpop(self.heap, (priority, heap_id, item))
        else:
            heapq.heappush(self.heap, (priority, heap_id, item))

        return len(self.heap)

    def get_items(self):
        """Return each item in the queue in order.

        Yields
        ------
        Any
            Item stored in the queue.

        """
        for priority, heap_id, item in self.heap:
            yield item


def get_probabilistic_group_sizes(groups, target_group_size, random_seed=None):
    """Create a dictionary of maximum sizes per group.

    Probabilistically generate varying sizes from a Poisson distribution. Make
    at least the given number of maximum attempts to generate sizes for which
    the total of all sizes is greater than zero.

    Examples
    --------
    Get sizes for two groups with a fractional maximum size. Their total
    size should still be an integer value greater than zero.

    >>> groups = ("2015", "2016")
    >>> seed = 314159
    >>> group_sizes = get_probabilistic_group_sizes(groups, 0.1, random_seed=seed)
    >>> int(sum(group_sizes.values())) > 0
    True

    A subsequent run of this function with the same groups and random seed
    should produce the same group sizes.

    >>> more_group_sizes = get_probabilistic_group_sizes(groups, 0.1, random_seed=seed)
    >>> list(group_sizes.values()) == list(more_group_sizes.values())
    True

    """
    assert target_group_size < 1.0

    # For small fractional maximum sizes, it is possible to randomly select
    # maximum queue sizes that all equal zero. When this happens, filtering
    # fails unexpectedly. We make multiple attempts to create queues with
    # maximum sizes greater than zero for at least one queue.
    random_generator = np.random.default_rng(random_seed)
    total_max_size = 0
    attempts = 0
    max_attempts = 100
    max_sizes_per_group = {}

    while total_max_size == 0 and attempts < max_attempts:
        for group in sorted(groups):
            max_sizes_per_group[group] = random_generator.poisson(target_group_size)

        total_max_size = sum(max_sizes_per_group.values())
        attempts += 1

    return max_sizes_per_group


TARGET_SIZE_COLUMN = '_augur_filter_target_size'
INPUT_SIZE_COLUMN = '_augur_filter_input_size'
OUTPUT_SIZE_COLUMN = '_augur_filter_subsampling_output_size'


def get_weighted_group_sizes(
        records_per_group: Dict[Group, int],
        group_by: List[str],
        weights_file: str,
        target_total_size: int,
        output_sizes_file: Optional[str],
        random_seed: Optional[int],
    ) -> Dict[Group, int]:
    """Return target group sizes based on weights defined in ``weights_file``.
    """
    groups = records_per_group.keys()

    weights = read_weights_file(weights_file)

    weighted_columns = get_weighted_columns(weights_file)

    # Other columns in group_by are considered unweighted.
    unweighted_columns = list(set(group_by) - set(weighted_columns))

    if unweighted_columns:
        # This has the side effect of weighting the values *alongside* (rather
        # than within) each weighted group. After dropping unused groups, adjust
        # weights to ensure equal weighting of unweighted columns *within* each
        # weighted group defined by the weighted columns.
        weights = _add_unweighted_columns(weights, groups, group_by, unweighted_columns)

        weights = _handle_incomplete_weights(weights, weights_file, weighted_columns, group_by, groups)
        weights = _drop_unused_groups(weights, groups, group_by)

        weights = _adjust_weights_for_unweighted_columns(weights, weighted_columns, unweighted_columns)
    else:
        weights = _handle_incomplete_weights(weights, weights_file, weighted_columns, group_by, groups)
        weights = _drop_unused_groups(weights, groups, group_by)

    weights = _calculate_weighted_group_sizes(weights, target_total_size, random_seed)

    # Add columns to summarize the input data
    weights[INPUT_SIZE_COLUMN] = weights.apply(lambda row: records_per_group[tuple(row[group_by].values)], axis=1)
    weights[OUTPUT_SIZE_COLUMN] = weights[[INPUT_SIZE_COLUMN, TARGET_SIZE_COLUMN]].min(axis=1)

    # Warn on any under-sampled groups
    for _, row in weights.iterrows():
        if row[INPUT_SIZE_COLUMN] < row[TARGET_SIZE_COLUMN]:
            sequences = _n('sequence', 'sequences', row[TARGET_SIZE_COLUMN])
            are = _n('is', 'are', row[INPUT_SIZE_COLUMN])
            group = list(f'{col}={value!r}' for col, value in row[group_by].items())
            print_err(f"WARNING: Targeted {row[TARGET_SIZE_COLUMN]} {sequences} for group {group} but only {row[INPUT_SIZE_COLUMN]} {are} available.")

    if output_sizes_file:
        weights.to_csv(output_sizes_file, index=False, sep='\t')

    return dict(zip(weights[group_by].apply(tuple, axis=1), weights[TARGET_SIZE_COLUMN]))


def _add_unweighted_columns(
        weights: pd.DataFrame,
        groups: Iterable[Group],
        group_by: List[str],
        unweighted_columns: List[str],
    ) -> pd.DataFrame:
    """Add the unweighted columns to the weights DataFrame.
    
    This is done by extending the existing weights to the newly created groups.
    """

    # Get unique values for each unweighted column.
    values_for_unweighted_columns = defaultdict(set)
    for group in groups:
        # NOTE: The ordering of entries in `group` corresponds to the column
        # names in `group_by`, but only because `get_groups_for_subsampling`
        # conveniently retains the order. This could be more tightly coupled,
        # but it works.
        column_to_value_map = dict(zip(group_by, group))
        for column in unweighted_columns:
            values_for_unweighted_columns[column].add(column_to_value_map[column])

    # Create a DataFrame for all permutations of values in unweighted columns.
    lists = [list(values_for_unweighted_columns[column]) for column in unweighted_columns]
    unweighted_permutations = pd.DataFrame(list(itertools.product(*lists)), columns=unweighted_columns)

    return pd.merge(unweighted_permutations, weights, how='cross')


def _drop_unused_groups(
        weights: pd.DataFrame,
        groups: Collection[Group],
        group_by: List[str],
    ) -> pd.DataFrame:
    """Drop any groups from ``weights`` that don't appear in ``groups``.
    """
    weights.set_index(group_by, inplace=True)

    # Pandas only uses MultiIndex if there is more than one column in the index.
    valid_index: Set[Union[Group, str]]
    if len(group_by) > 1:
        valid_index = set(groups)
    else:
        valid_index = set(group[0] for group in groups)

    extra_groups = set(weights.index) - valid_index
    if extra_groups:
        count = len(extra_groups)
        unit = _n("group", "groups", count)
        print_err(f"NOTE: Skipping {count} {unit} due to lack of entries in metadata.")
        weights = weights[weights.index.isin(valid_index)]

    weights.reset_index(inplace=True)

    return weights


def _adjust_weights_for_unweighted_columns(
        weights: pd.DataFrame,
        weighted_columns: List[str],
        unweighted_columns: Collection[str],
    ) -> pd.DataFrame:
    """Adjust weights for unweighted columns to reflect equal weighting within each weighted group.
    """
    columns = _n('column', 'columns', len(unweighted_columns))
    those = _n('that', 'those', len(unweighted_columns))
    print_err(f"NOTE: Weights were not provided for the {columns} {', '.join(repr(col) for col in unweighted_columns)}. Using equal weights across values in {those} {columns}.")        

    weights_grouped = weights.groupby(weighted_columns)
    weights[WEIGHTS_COLUMN] = weights_grouped[WEIGHTS_COLUMN].transform(lambda x: x / len(x))

    return weights


def _calculate_weighted_group_sizes(
        weights: pd.DataFrame,
        target_total_size: int,
        random_seed: Optional[int],
    ) -> pd.DataFrame:
    """Calculate maximum group sizes based on weights.
    """
    weights[TARGET_SIZE_COLUMN] = pd.Series(weights[WEIGHTS_COLUMN] / weights[WEIGHTS_COLUMN].sum() * target_total_size)

    # Group sizes must be whole numbers. Round probabilistically by adding a
    # random number between [0,1) and truncating the decimal part.
    rng = np.random.default_rng(random_seed)
    weights[TARGET_SIZE_COLUMN] = (weights[TARGET_SIZE_COLUMN].add(pd.Series(rng.random(len(weights))))).astype(int)

    return weights


def _handle_incomplete_weights(
        weights: pd.DataFrame,
        weights_file: str,
        weighted_columns: List[str],
        group_by: List[str],
        groups: Iterable[Group],
    ) -> pd.DataFrame:
    """Handle the case where the weights file does not cover all rows in the metadata.
    """
    missing_groups = set(groups) - set(weights[group_by].apply(tuple, axis=1))

    if not missing_groups:
        return weights

    # Collect the column values that are missing weights.
    missing_values_by_column = defaultdict(set)
    for group in missing_groups:
        # NOTE: The ordering of entries in `group` corresponds to the column
        # names in `group_by`, but only because `get_groups_for_subsampling`
        # conveniently retains the order. This could be more tightly coupled,
        # but it works.
        column_to_value_map = dict(zip(group_by, group))
        for column in weighted_columns:
            missing_values_by_column[column].add(column_to_value_map[column])

    columns_with_values = '\n            - '.join(f'{column!r}: {list(sorted(values))}' for column, values in sorted(missing_values_by_column.items()))

    default_weight = get_default_weight(weights, weighted_columns)

    if not default_weight:
        raise AugurError(dedent(f"""\
            The input metadata contains these values under the following columns that are not covered by {weights_file!r}:
            - {columns_with_values}
            To fix this, either:
            (1) specify weights explicitly - add entries to {weights_file!r} for the values above, or
            (2) specify a default weight - add an entry to {weights_file!r} with the value {COLUMN_VALUE_FOR_DEFAULT_WEIGHT!r} for all columns"""))
    else:
        print_err(dedent(f"""\
            WARNING: The input metadata contains these values under the following columns that are not directly covered by {weights_file!r}:
            - {columns_with_values}
            The default weight of {default_weight!r} will be used for all groups defined by those values."""))

        missing_weights = pd.DataFrame(sorted(missing_groups), columns=group_by)
        missing_weights[WEIGHTS_COLUMN] = default_weight
        return pd.merge(weights, missing_weights, on=[*group_by, WEIGHTS_COLUMN], how='outer')


def create_queues_by_group(max_sizes_per_group):
    return {group: PriorityQueue(max_size)
            for group, max_size in max_sizes_per_group.items()}


def calculate_sequences_per_group(target_max_value, group_sizes, allow_probabilistic=True):
    """Calculate the number of sequences per group for a given maximum number of
    sequences to be returned and the number of sequences in each requested
    group. Optionally, allow the result to be probabilistic such that the mean
    result of a Poisson process achieves the calculated sequences per group for
    the given maximum.

    Parameters
    ----------
    target_max_value : int
        Maximum number of sequences to return by subsampling at some calculated
        number of sequences per group for the given counts per group.
    group_sizes : list of int
        A list with the number of sequences in each requested group.
    allow_probabilistic : bool
        Whether to allow probabilistic subsampling when the number of groups
        exceeds the requested maximum.

    Raises
    ------
    TooManyGroupsError
        When there are more groups than sequences per group and probabilistic
        subsampling is not allowed.

    Returns
    -------
    int or float :
        Number of sequences per group.
    bool :
        Whether probabilistic subsampling was used.

    """
    probabilistic_used = False

    try:
        sequences_per_group = _calculate_sequences_per_group(
            target_max_value,
            group_sizes,
        )
    except TooManyGroupsError as error:
        if allow_probabilistic:
            print_err(f"WARNING: {error}")
            sequences_per_group = target_max_value / len(group_sizes)
            probabilistic_used = True
        else:
            raise error

    return sequences_per_group, probabilistic_used


class TooManyGroupsError(ValueError):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return str(self.msg)


def _calculate_total_sequences(
        hypothetical_spg: float, group_sizes: Collection[int],
) -> float:
    # calculate how many sequences we'd keep given a hypothetical spg.
    return sum(
        min(hypothetical_spg, group_count)
        for group_count in group_sizes
    )


def _calculate_sequences_per_group(
        target_max_value: int,
        group_sizes: Collection[int]
) -> int:
    """This is partially inspired by
    https://github.com/python/cpython/blob/3.8/Lib/bisect.py

    This should return the spg such that we don't exceed the requested
    number of samples.

    Parameters
    ----------
    target_max_value : int
        the total number of sequences allowed across all groups
    group_sizes : Collection[int]
        the number of sequences in each group

    Returns
    -------
    int
        maximum number of sequences allowed per group to meet the required maximum total
        sequences allowed

    Examples
    --------
    >>> _calculate_sequences_per_group(4, [4, 2])
    2
    >>> _calculate_sequences_per_group(2, [4, 2])
    1
    >>> _calculate_sequences_per_group(1, [4, 2])
    Traceback (most recent call last):
        ...
    augur.filter.subsample.TooManyGroupsError: Asked to provide at most 1 sequences, but there are 2 groups.
    """

    if len(group_sizes) > target_max_value:
        # we have more groups than sequences we are allowed, which is an
        # error.

        raise TooManyGroupsError(
            "Asked to provide at most {} sequences, but there are {} "
            "groups.".format(target_max_value, len(group_sizes)))

    lo = 1
    hi = target_max_value

    while hi - lo > 2:
        mid = (hi + lo) // 2
        if _calculate_total_sequences(mid, group_sizes) <= target_max_value:
            lo = mid
        else:
            hi = mid

    if _calculate_total_sequences(hi, group_sizes) <= target_max_value:
        return int(hi)
    else:
        return int(lo)



import json
import jsonschema.validators
import pytest
from pathlib import Path

schemas = list(Path("augur/data/").glob("schema-*.json"))

@pytest.mark.parametrize("schema_path", schemas, ids = lambda schema_path: str(schema_path))
def test_schema_is_valid(schema_path):
    with schema_path.open("rb") as schema_fh:
        schema = json.load(schema_fh)

    Validator = jsonschema.validators.validator_for(schema)
    Validator.check_schema(schema)



from unittest import mock
from augur.export_v2 import is_numeric, format_number, attr_confidence

class TestIsNumeric():
    """This tests _our_ is_numeric function, it's not a philosophical discussion about numbers!"""
    def test_strings_are_not_numeric(self):
        assert not is_numeric("abc")
        assert not is_numeric("1")
        assert not is_numeric("1.0")
    def test_numbers_are_numeric(self):
        assert is_numeric(1)
        assert is_numeric(1.1)
        assert is_numeric(1.04e-2)

class TestFormatNumber():
    def test_integers_are_unchanged(self):
        assert 123 == format_number(123)
        assert 0 == format_number(0)
        assert -123 == format_number(-123)

    def test_positive_float_sig_figs(self):
        assert "2010.0" == str(format_number(2010.0))
        assert "2010.0" == str(format_number(2010.0000000001))
        assert "2010.123" == str(format_number(2010.1234000000001))
        assert "2010.0" == str(format_number(2010.0001))
        assert "0.123" == str(format_number(0.123456))

    def test_negative_float_sig_figs(self):
        assert "-2010.0" == str(format_number(-2010.0))
        assert "-2010.0" == str(format_number(-2010.0000000001))
        assert "-2010.123" == str(format_number(-2010.1234000000001))
        assert "-2010.0" == str(format_number(-2010.0001))
        assert "-0.123" == str(format_number(-0.123456))

    def test_exponential_notation_sig_figs(self):
        assert "-1e-12" == str(format_number(-1.000088900581841e-12))
        assert "-1.01e-12" == str(format_number(-1.0088900581841e-12))
        assert "-8.89e-15" == str(format_number(-0.0088900581841e-12))
        
        assert "1e-12" == str(format_number(1.000088900581841e-12))
        assert "1.01e-12" == str(format_number(1.0088900581841e-12))
        assert "8.89e-15" == str(format_number(0.0088900581841e-12))

    def test_exponential_notation_conversion(self):
        assert ("1.23e-10") == str(format_number(0.000_000_000_123_456))


def no_op(*args):
    pass

class TestConfidenceExtraction():
    def test_no_confidence(self):
        assert {} == attr_confidence("name", {"key": "something"}, "key")

    def test_no_confidence_but_entropy(self):
        assert {} == attr_confidence("name", {"key": "something", "key_entropy": 2.0}, "key")

    @mock.patch("augur.export_v2.warn", no_op)
    def test_confidence_but_no_entropy(self):
        # array confidences don't need entropy (and are represented internally as tuples)
        assert {'confidence': (1,2)} == attr_confidence("name", {"key": 1.5, "key_confidence": [1,2]}, "key")
        # but dict confidences do...
        assert {} == attr_confidence("name", {"key": "foo", "key_confidence": {"foo": 0.99}}, "key")

    @mock.patch("augur.export_v2.warn", no_op)
    def test_array_confidence(self):
        # Input JSON array -> python tuple -> exported as JSON Array
        assert {'confidence': (1,2)} == attr_confidence("name", {"key": 1.5, "key_confidence": [1,2]}, "key")
        # invalid confidence. Warnings suppressed.
        assert {} == attr_confidence("name", {"key": 1.5, "key_confidence": [1,2,3]}, "key")


    @mock.patch("augur.export_v2.warn", no_op)
    def test_dict_confidence(self):
        conf = {"x": 0.888, "y": 1.1e-2}
        entropy = 0.123
        assert {'confidence': conf, "entropy": entropy} == \
            attr_confidence("name", {"key": "x", "key_confidence": conf, "key_entropy": entropy}, "key")

        # values must all be numbers
        assert {} == \
            attr_confidence("name", {"key": "x", "key_confidence": {**conf, "c": "d"}, "key_entropy": entropy}, "key")    

    @mock.patch("augur.export_v2.warn", no_op)
    def test_unknown_confidence(self):
        assert {} == attr_confidence("name", {"key": "x", "key_confidence": "0.999"}, "key")


import pandas as pd
from textwrap import dedent
from typing import List
from augur.errors import AugurError


WEIGHTS_COLUMN = 'weight'
COLUMN_VALUE_FOR_DEFAULT_WEIGHT = 'default'


class InvalidWeightsFile(AugurError):
    def __init__(self, file, error_message):
        super().__init__(f"Bad weights file {file!r}.\n{error_message}")


def read_weights_file(weights_file):
    weights = pd.read_csv(weights_file, delimiter='\t', comment='#')

    if not pd.api.types.is_numeric_dtype(weights[WEIGHTS_COLUMN]):
        non_numeric_weight_lines = [index + 2 for index in weights[~weights[WEIGHTS_COLUMN].str.isnumeric()].index.tolist()]
        raise InvalidWeightsFile(weights_file, dedent(f"""\
            Found non-numeric weights on the following lines: {non_numeric_weight_lines}
            {WEIGHTS_COLUMN!r} column must be numeric."""))

    if any(weights[WEIGHTS_COLUMN] < 0):
        negative_weight_lines = [index + 2 for index in weights[weights[WEIGHTS_COLUMN] < 0].index.tolist()]
        raise InvalidWeightsFile(weights_file, dedent(f"""\
            Found negative weights on the following lines: {negative_weight_lines}
            {WEIGHTS_COLUMN!r} column must be non-negative."""))

    return weights


def get_weighted_columns(weights_file):
    with open(weights_file) as f:
        has_rows = False
        for row in f:
            has_rows = True
            if row.startswith('#'):
                continue
            columns = row.rstrip().split('\t')
            break
    if not has_rows:
        raise InvalidWeightsFile(weights_file, "File is empty.")
    columns.remove(WEIGHTS_COLUMN)
    return columns


def get_default_weight(weights: pd.DataFrame, weighted_columns: List[str]):
    default_weight_values = weights[(weights[weighted_columns] == COLUMN_VALUE_FOR_DEFAULT_WEIGHT).all(axis=1)][WEIGHTS_COLUMN].unique()

    if len(default_weight_values) > 1:
        # TODO: raise InvalidWeightsFile, not AugurError. This function takes
        # the weights DataFrame instead of the filepath, so it does not have the
        # file parameter to InvalidWeightsFile. I didn't want to pass an extra
        # filepath parameter to this function just to have it available for the
        # custom exception class. I also didn't want to pass the filepath
        # parameter. One idea would be to define a custom class to represent a
        # weights file, however this seemed overkill in a quick prototype:
        # <https://github.com/nextstrain/augur/commit/08408580a6beed91e46e93cee28d2ebd552bdbba>
        raise AugurError(f"Multiple default weights were specified: {', '.join(repr(weight) for weight in default_weight_values)}. Only one default weight entry can be accepted.")
    if len(default_weight_values) == 1:
        return default_weight_values[0]



import gettext
import sys

from augur.debug import DEBUGGING


def print_err(*args):
    """Print to stderr. When data goes to stdout (most cases), this should be
    used for any informational messages, not just errors/warnings."""
    print(*args, file=sys.stderr)


def print_debug(*args):
    """Print to stderr if in debugging mode."""
    if DEBUGGING:
        print_err(*args)


# Use ngettext() without a message catalog for its singular/plural handling so
# we can make proper error messages.  gettext() (no "n") is conventionally
# aliased as "_", so alias ngettext() as "_n".
_n = gettext.NullTranslations().ngettext



"""
Print the version of augur.
"""
from .__version__ import __version__

def register_parser(parent_subparsers):
    return parent_subparsers.add_parser("version", help=__doc__)

def run(args):
    print("augur", __version__)
    return 0



r"""
Read a file like Augur, with transparent optimized decompression and universal newlines.

Input is read from the given file path, as the compression format detection
requires a seekable stream.  The given path may be "-" to explicitly read from
stdin, but no decompression will be done.

Output is always to stdout.

Universal newline translation is always performed, so \n, \r\n, and \r in the
input are all translated to the system's native newlines (e.g. \n on Unix, \r\n
on Windows) in the output.
"""
import io
import os
import signal
import sys
from shutil import copyfileobj

from .io.file import open_file
from .utils import first_line


# The buffer size used by xopen() (which underlies open_file()), which notes:
# 128KB [KiB] buffer size also used by cat, pigz etc. It is faster than the 8K
# [KiB] default.
BUFFER_SIZE = max(io.DEFAULT_BUFFER_SIZE, 128 * 1024)

SIGPIPE = getattr(signal, "SIGPIPE", None)


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("read-file", help=first_line(__doc__))
    parser.add_argument("path", metavar="PATH", help="path to file")
    return parser


def run(args):
    with open_file(args.path, "rt", newline=None) as f:
        # It's tempting to want to splice(2) here, but it turns out to make
        # little sense.  Firstly, the availability of splice(2) is Linux,
        # Python ≥3.10, and one of the files needs to be a pipe.  The chance of
        # all of those together is slim-to-none, particularly because even in
        # the common case of xopen() reading from a pipe—the stdout of an
        # external decompression process—we can't use that pipe directly
        # because xopen() always buffers the first block of the file into
        # Python.¹  Secondly, we want universal newline handling—so that
        # callers get behaviour consistent with the rest of Augur—and that
        # rules out splice(2).
        #
        # Copying the data thru Python instead of with splice(2) seems fast
        # enough in some quick trials with large files (e.g. against `zstd`
        # directly), and the bottlenecks in pipelines using this command will
        # often not be this command's i/o.
        #   -trs, 11 July 2024, updated 24 July 2024
        #
        # ¹ <https://github.com/pycompression/xopen/blob/67651844/src/xopen/__init__.py#L374>

        # Handle SIGPIPE, which Python converts to BrokenPipeError, gracefully
        # and like most Unix programs.  See also
        # <https://docs.python.org/3/library/signal.html#note-on-sigpipe>.
        try:
            copyfileobj(f, sys.stdout, BUFFER_SIZE)

            # Force a flush so if SIGPIPE is going to happen it happens now.
            sys.stdout.flush()
        except BrokenPipeError:
            # Avoid errors from Python automatically flushing stdout on exit.
            devnull = os.open(os.devnull, os.O_WRONLY)
            os.dup2(devnull, sys.stdout.fileno())

            # Return conventional exit status for "killed by SIGPIPE" on Unix.
            return 128 + SIGPIPE if SIGPIPE else 1



"""
Reconstruct alignments from mutations inferred on the tree
"""

from Bio import SeqIO, Seq, SeqRecord, Phylo
from .io.file import open_file
from .utils import read_node_data



def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("reconstruct-sequences", help=__doc__)
    parser.add_argument('--tree', required=True, help="tree as Newick file")
    parser.add_argument('--gene', type=str, help="gene to translate (list or file containing list)")
    parser.add_argument('--mutations', required=True, type=str, help="json file containing mutations "
                            "mapped to each branch and the sequence of the root.")
    parser.add_argument('--vcf-aa-reference', type=str, help='fasta file of the reference gene translations for VCF format')
    parser.add_argument('--internal-nodes', action='store_true', help="include sequences of internal nodes in output")
    parser.add_argument('--output', type=str)
    return parser


def get_sequence(pseq, muts):
    """reconstruct a child sequence from that of its parent and the mutations

    Parameters
    ----------
    pseq : str
        sequence of the parent as string, won't be changed!
    muts : list
        mutations of the type K135T where K is the parental state, 135 is the
        position in 1-based numbering, and T is the new state

    Returns
    -------
    str
        reconstructed sequence
    """
    pseq_list = list(pseq)
    for mut in muts:
        new_state = mut[-1]
        pos = int(mut[1:-1])-1
        assert pseq_list[pos]==mut[0]
        pseq_list[pos]=new_state

    return "".join(pseq_list)


def load_alignments(sequence_files, gene_names):
    from Bio import AlignIO
    alignments = {}
    for fname, gene in zip(sequence_files, gene_names):
        alignments[gene] = AlignIO.read(fname, 'fasta')
    return alignments


def run(args):
    ## read tree and data, if reading data fails, return with error code
    tree = Phylo.read(args.tree, 'newick')

    #check whether VCF - by existance of VCF reference file
    is_vcf = args.vcf_aa_reference is not None

    ## check file format and read in sequences
    node_data = read_node_data(args.mutations, args.tree)
    if node_data is None:
        print("ERROR: could not read mutation data "+ ("(incl sequences)" if not is_vcf else ""))
        return 1

    root_node = tree.root.name

    #if VCF, read in the reference seq for each gene, put on root
    if(is_vcf):
        node_data["nodes"][root_node]['aa_sequences'] = {}
        with open_file(args.vcf_aa_reference) as handle:
            for record in SeqIO.parse(handle, "fasta"):
                if record.id==args.gene:
                    #'root' may not be same as 'reference', so apply any mutations at root here!
                    node_data["nodes"][root_node]['aa_sequences'][record.id] = get_sequence(str(record.seq), node_data["nodes"][root_node]["aa_muts"][record.id])

    # check that root node has sequences for each requested gene
    if "aa_sequences" not in node_data["nodes"][root_node]:
        print("ERROR: ancestral sequences are not provided")
        return 1
    if not args.gene in node_data["nodes"][root_node]['aa_sequences']:
        print("ERROR: ancestral sequences missing for gene",args.gene)
        return 1

    # gather all reconstructed sequences
    sequences = dict()
    is_terminal = {}

    sequences[root_node] = node_data["nodes"][root_node]['aa_sequences'][args.gene]
    is_terminal[root_node] = False
    for node in tree.get_nonterminals(order='preorder'):
        parent_sequence = sequences[node.name]
        for child in node:
            sequences[child.name] = get_sequence(parent_sequence, node_data["nodes"][child.name]["aa_muts"][args.gene])
            is_terminal[child.name] = child.is_terminal()

    # write alignments to file
    seqs = [SeqRecord.SeqRecord(seq=Seq.Seq(sequences[strain]), id=strain, name=strain, description='')
            for strain in sequences if is_terminal[strain] or args.internal_nodes]
    SeqIO.write(seqs, args.output, 'fasta')



"""
Applies user curated geolocation rules to the geolocation fields.
"""
from collections import defaultdict
from augur.errors import AugurError
from augur.io.print import print_err
from augur.utils import first_line


class CyclicGeolocationRulesError(AugurError):
    pass


def load_geolocation_rules(geolocation_rules_file):
    """
    Loads the geolocation rules from the provided *geolocation_rules_file*.
    Returns the rules as a dict:

.. code-block:: text

    {
        regions: {
            countries: {
                divisions: {
                    locations: corrected_geolocations_tuple
                }
            }
        }
    }
    """
    geolocation_rules = defaultdict(lambda: defaultdict(lambda: defaultdict(dict)))
    with open(geolocation_rules_file, 'r') as rules_fh:
        for line in rules_fh:
            # ignore comments
            if line.strip()=="" or line.lstrip()[0] == '#':
                continue

            row = line.strip('\n').split('\t')
            # Skip lines that cannot be split into raw and annotated geolocations
            if len(row) != 2:
                print_err(
                    f"WARNING: Could not decode geolocation rule {line!r}.",
                    "Please make sure rules are formatted as",
                    "'region/country/division/location<tab>region/country/division/location'.")
                continue

            # remove trailing comments
            row[-1] = row[-1].partition('#')[0].rstrip()
            raw , annot = tuple( row[0].split('/') ) , tuple( row[1].split('/') )

            # Skip lines where raw or annotated geolocations cannot be split into 4 fields
            if len(raw) != 4:
                print_err(
                    f"WARNING: Could not decode the raw geolocation {row[0]!r}.",
                    "Please make sure it is formatted as 'region/country/division/location'.")
                continue

            if len(annot) != 4:
                print_err(
                    f"WARNING: Could not decode the annotated geolocation {row[1]!r}.",
                    "Please make sure it is formatted as 'region/country/division/location'.")
                continue


            geolocation_rules[raw[0]][raw[1]][raw[2]][raw[3]] = annot

    return geolocation_rules


def get_annotated_geolocation(geolocation_rules, raw_geolocation, rule_traversal = None):
    """
    Gets the annotated geolocation for the *raw_geolocation* in the provided
    *geolocation_rules*.

    Recursively traverses the *geolocation_rules* until we get the annotated
    geolocation, which must be a Tuple. Returns `None` if there are no
    applicable rules for the provided *raw_geolocation*.

    Rules are applied in the order of region, country, division, location.
    First checks the provided raw values for geolocation fields, then if there
    are not matches, tries to use general rules marked with '*'.
    """
    # Always instantiate the rule traversal as an empty list if not provided,
    # e.g. the first call of this recursive function
    if rule_traversal is None:
        rule_traversal = []

    current_rules = geolocation_rules
    # Traverse the geolocation rules based using the rule_traversal values
    for field_value in rule_traversal:
        current_rules = current_rules.get(field_value)
        # If we hit `None`, then we know there are no matching rules, so stop the rule traversal
        if current_rules is None:
            break

    # We've found the tuple of the annotated geolocation
    if isinstance(current_rules, tuple):
        return current_rules

    # We've reach the next level of geolocation rules,
    # so try to traverse the rules with the next target in raw_geolocation
    if isinstance(current_rules, dict):
        next_traversal_target = raw_geolocation[len(rule_traversal)]
        rule_traversal.append(next_traversal_target)
        return get_annotated_geolocation(geolocation_rules, raw_geolocation, rule_traversal)

    # We did not find any matching rule for the last traversal target
    if current_rules is None:
        # If we've used all general rules and we still haven't found a match,
        # then there are no applicable rules for this geolocation
        if all(value == '*' for value in rule_traversal):
            return None

        # If we failed to find matching rule with a general rule as the last
        # traversal target, then delete all trailing '*'s to reset rule_traversal
        # to end with the last index that is currently NOT a '*'
        # [A, *, B, *] => [A, *, B]
        # [A, B, *, *] => [A, B]
        # [A, *, *, *] => [A]
        if rule_traversal[-1] == '*':
            # Find the index of the first of the consecutive '*' from the
            # end of the rule_traversal
            # [A, *, B, *] => first_consecutive_general_rule_index = 3
            # [A, B, *, *] => first_consecutive_general_rule_index = 2
            # [A, *, *, *] => first_consecutive_general_rule_index = 1
            for index, field_value in reversed(list(enumerate(rule_traversal))):
                if field_value == '*':
                    first_consecutive_general_rule_index = index
                else:
                    break

            rule_traversal = rule_traversal[:first_consecutive_general_rule_index]

        # Set the final value to '*' in hopes that by moving to a general rule,
        # we can find a matching rule.
        rule_traversal[-1] = '*'

        return get_annotated_geolocation(geolocation_rules, raw_geolocation, rule_traversal)


def transform_geolocations(geolocation_rules, geolocation):
    """
    Transform the provided *geolocation* by looking it up in the provided
    *geolocation_rules*.

    This will use all rules that apply to the geolocation and rules will
    be applied in the order of region, country, division, location.

    Returns the original geolocation if no geolocation rules apply.

    Raises a `CyclicGeolocationRulesError` if more than 1000 rules have
    been applied to the raw geolocation.
    """
    transformed_values = geolocation
    rules_applied = 0
    continue_to_apply = True

    while continue_to_apply:
        annotated_values = get_annotated_geolocation(geolocation_rules, transformed_values)

        # Stop applying rules if no annotated values were found
        if annotated_values is None:
            continue_to_apply = False
        else:
            rules_applied += 1

            if rules_applied > 1000:
                raise CyclicGeolocationRulesError(
                    f"More than 1000 geolocation rules applied on the same entry {geolocation!r}."
                )

            # Create a new list of values for comparison to previous values
            new_values = list(transformed_values)
            for index, value in enumerate(annotated_values):
                # Keep original value if annotated value is '*'
                if value != '*':
                    new_values[index] = value

            # Stop applying rules if this rule did not change the values,
            # since this means we've reach rules with '*' that no longer change values
            if new_values == transformed_values:
                continue_to_apply = False

            transformed_values = new_values

    return transformed_values


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("apply-geolocation-rules",
        parents=[parent_subparsers.shared_parser],
        help=first_line(__doc__))

    parser.add_argument("--region-field", default="region",
        help="Field that contains regions in NDJSON records.")
    parser.add_argument("--country-field", default="country",
        help="Field that contains countries in NDJSON records.")
    parser.add_argument("--division-field", default="division",
        help="Field that contains divisions in NDJSON records.")
    parser.add_argument("--location-field", default="location",
        help="Field that contains location in NDJSON records.")
    parser.add_argument("--geolocation-rules", metavar="TSV", required=True,
        help="TSV file of geolocation rules with the format: " +
             "'<raw_geolocation><tab><annotated_geolocation>' where the raw and annotated geolocations " +
             "are formatted as '<region>/<country>/<division>/<location>'. " +
             "If creating a general rule, then the raw field value can be substituted with '*'." +
             "Lines starting with '#' will be ignored as comments." +
             "Trailing '#' will be ignored as comments.")

    return parser


def run(args, records):
    location_fields = [args.region_field, args.country_field, args.division_field, args.location_field]

    geolocation_rules = load_geolocation_rules(args.geolocation_rules)

    for record in records:

        annotated_values = transform_geolocations(geolocation_rules, [record.get(field, '') for field in location_fields])

        for index, field in enumerate(location_fields):
            record[field] = annotated_values[index]

        yield record



"""
Merge two or more metadata tables into one.

Tables must be given unique names to identify them in the output and are
merged in the order given.

Rows are joined by id (e.g. "strain" or "name" or other
--metadata-id-columns), and ids must be unique within an input table (i.e.
tables cannot contain duplicate ids).  All rows are output, even if they
appear in only a single table (i.e. a full outer join in SQL terms).

Columns are combined by name, either extending the combined table with a new
column or overwriting values in an existing column.  For columns appearing in
more than one table, non-empty values on the right hand side overwrite values
on the left hand side.  The first table's id column name is used as the output
id column name.  Non-id columns in other input tables that would conflict with
this output id column name are not allowed and if present will cause an error.

One generated column per input table may be optionally appended to the end of
the output table to identify the source of each row's data.  Column names are
generated with the template given to --source-columns where "{NAME}" in the
template is replaced by the table name given to --metadata.  Values in each
column are 1 or 0 for present or absent in that input table.  By default no
source columns are generated.  You may make this behaviour explicit with
--no-source-columns.

Metadata tables of arbitrary size can be handled, limited only by available
disk space.  Tables are not required to be entirely loadable into memory.  The
transient disk space required is approximately the sum of the uncompressed size
of the inputs.

SQLite is used behind the scenes to implement the merge, but, at least for now,
this should be considered an implementation detail that may change in the
future.  The SQLite 3 CLI, sqlite3, must be available.  If it's not on PATH (or
you want to use a version different from what's on PATH), set the SQLITE3
environment variable to path of the desired sqlite3 executable.
"""
import os
import re
import subprocess
import sys
from functools import reduce
from itertools import starmap
from shlex import quote as shquote
from shutil import which
from tempfile import mkstemp
from textwrap import dedent
from typing import Iterable, Tuple, TypeVar

from augur.argparse_ import ExtendOverwriteDefault, SKIP_AUTO_DEFAULT_IN_HELP
from augur.errors import AugurError
from augur.io.metadata import DEFAULT_DELIMITERS, DEFAULT_ID_COLUMNS, Metadata
from augur.io.print import print_err, print_debug, _n
from augur.utils import first_line


T = TypeVar('T')


class NamedMetadata(Metadata):
    name: str
    """User-provided descriptive name for this metadata file."""

    table_name: str
    """Generated SQLite table name for this metadata file, based on *name*."""

    def __init__(self, name: str, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.name = name
        self.table_name = f"metadata_{self.name}"

    def __repr__(self):
        return f"<NamedMetadata {self.name}={self.path}>"


def register_parser(parent_subparsers):
    parser = parent_subparsers.add_parser("merge", help=first_line(__doc__))

    input_group = parser.add_argument_group("inputs", "options related to input")
    input_group.add_argument("--metadata", nargs="+", action="extend", required=True, metavar="NAME=FILE", help="Required. Metadata table names and file paths. Names are arbitrary monikers used solely for referring to the associated input file in other arguments and in output column names. Paths must be to seekable files, not unseekable streams. Compressed files are supported." + SKIP_AUTO_DEFAULT_IN_HELP)

    input_group.add_argument("--metadata-id-columns", default=DEFAULT_ID_COLUMNS, nargs="+", action=ExtendOverwriteDefault, metavar="[TABLE=]COLUMN", help=f"Possible metadata column names containing identifiers, considered in the order given. Columns will be considered for all metadata tables by default. Table-specific column names may be given using the same names assigned in --metadata. Only one ID column will be inferred for each table. (default: {' '.join(map(shquote_humanized, DEFAULT_ID_COLUMNS))})" + SKIP_AUTO_DEFAULT_IN_HELP)
    input_group.add_argument("--metadata-delimiters", default=DEFAULT_DELIMITERS, nargs="+", action=ExtendOverwriteDefault, metavar="[TABLE=]CHARACTER", help=f"Possible field delimiters to use for reading metadata tables, considered in the order given. Delimiters will be considered for all metadata tables by default. Table-specific delimiters may be given using the same names assigned in --metadata. Only one delimiter will be inferred for each table. (default: {' '.join(map(shquote_humanized, DEFAULT_DELIMITERS))})" + SKIP_AUTO_DEFAULT_IN_HELP)

    output_group = parser.add_argument_group("outputs", "options related to output")
    output_group.add_argument('--output-metadata', required=True, metavar="FILE", help="Required. Merged metadata as TSV. Compressed files are supported." + SKIP_AUTO_DEFAULT_IN_HELP)
    output_group.add_argument('--source-columns', metavar="TEMPLATE", help=f"Template with which to generate names for the columns (described above) identifying the source of each row's data. Must contain a literal placeholder, {{NAME}}, which stands in for the metadata table names assigned in --metadata. (default: disabled)" + SKIP_AUTO_DEFAULT_IN_HELP)
    output_group.add_argument('--no-source-columns', dest="source_columns", action="store_const", const=None, help=f"Suppress generated columns (described above) identifying the source of each row's data. This is the default behaviour, but it may be made explicit or used to override a previous --source-columns." + SKIP_AUTO_DEFAULT_IN_HELP)
    output_group.add_argument('--quiet', action="store_true", default=False, help="Suppress informational and warning messages normally written to stderr. (default: disabled)" + SKIP_AUTO_DEFAULT_IN_HELP)

    return parser


def run(args):
    print_info = print_err if not args.quiet else lambda *_: None

    # Parse --metadata arguments
    if not len(args.metadata) >= 2:
        raise AugurError(f"At least two metadata inputs are required for merging.")

    if unnamed := [repr(x) for x in args.metadata if "=" not in x or x.startswith("=")]:
        raise AugurError(dedent(f"""\
            All metadata inputs must be assigned a name, e.g. with NAME=FILE.

            The following {_n("input was", "inputs were", len(unnamed))} missing a name:

              {indented_list(unnamed, '            ' + '  ')}
            """))

    metadata = pairs(args.metadata)

    if duplicate_names := [repr(name) for name, count
                                       in count_unique(name for name, _ in metadata)
                                       if count > 1]:
        raise AugurError(dedent(f"""\
            Metadata input names must be unique.

            The following {_n("name was", "names were", len(duplicate_names))} used more than once:

              {indented_list(duplicate_names, '            ' + '  ')}
            """))


    # Parse --metadata-id-columns and --metadata-delimiters
    metadata_names = set(name for name, _ in metadata)

    metadata_id_columns = pairs(args.metadata_id_columns)
    metadata_delimiters = pairs(args.metadata_delimiters)

    if unknown_names := [repr(name) for name, _ in metadata_id_columns if name and name not in metadata_names]:
        raise AugurError(dedent(f"""\
            Unknown metadata table {_n("name", "names", len(unknown_names))} in --metadata-id-columns:

              {indented_list(unknown_names, '            ' + '  ')}

            {_n("This name does", "These names do", len(unknown_names))} not appear in the NAME=FILE pairs given to --metadata.
            """))

    if unknown_names := [repr(name) for name, _ in metadata_delimiters if name and name not in metadata_names]:
        raise AugurError(dedent(f"""\
            Unknown metadata table {_n("name", "names", len(unknown_names))} in --metadata-delimiters:

              {indented_list(unknown_names, '            ' + '  ')}

            {_n("This name does", "These names do", len(unknown_names))} not appear in the NAME=FILE pairs given to --metadata.
            """))


    # Validate --source-columns template and convert to template function
    output_source_column = None

    if args.source_columns is not None:
        if "{NAME}" not in args.source_columns:
            raise AugurError(dedent(f"""\
                The --source-columns template must contain the literal
                placeholder {{NAME}} but the given value ({args.source_columns!r}) does not.

                You may need to quote the whole template value to prevent your
                shell from interpreting the placeholder before Augur sees it.
                """))

        output_source_column = lambda name: args.source_columns.replace('{NAME}', name)


    # Infer delimiters and id columns
    metadata = [
        NamedMetadata(name, path, [delim  for name_, delim  in metadata_delimiters if not name_ or name_ == name] or DEFAULT_DELIMITERS,
                                  [column for name_, column in metadata_id_columns if not name_ or name_ == name] or DEFAULT_ID_COLUMNS)
            for name, path in metadata]


    # Locate how to re-invoke ourselves (_this_ specific Augur).
    if sys.executable:
        augur = f"{shquote(sys.executable)} -m augur"
    else:
        # A bit unusual we don't know our own Python executable, but assume we
        # can access ourselves as the ``augur`` command.
        augur = f"augur"


    # Work with a temporary, on-disk SQLite database under a name we control so
    # we can access it from multiple (serial) processes.
    db_fd, db_path = mkstemp(prefix="augur-merge-", suffix=".sqlite")
    os.close(db_fd)

    # Clean up database file by default
    delete_db = True

    # Track columns as we see them, in order.  The first metadata's id column
    # is always the first output column of the merge, so insert it now.
    output_id_column = metadata[0].id_column
    output_columns = { output_id_column: [] }

    if conflicting_columns := [f"{c!r} in metadata table {m.name!r} (id column: {m.id_column!r})"
                                    for m in metadata
                                    for c in m.columns
                                     if c == output_id_column
                                    and c != m.id_column]:
        raise AugurError(dedent(f"""\
            Non-id column names in metadata inputs may not conflict with the
            output id column name ({output_id_column!r}, the first input's id column).

            The following input {_n("column", "columns", len(conflicting_columns))} would conflict:

              {indented_list(conflicting_columns, '            ' + '  ')}

            Please rename or drop the conflicting {_n("column", "columns", len(conflicting_columns))} before merging.
            Renaming may be done with `augur curate rename`.
            """))

    output_source_columns = set(
        output_source_column(m.name)
            for m in metadata
             if output_source_column)

    if conflicting_columns := [f"{c!r} in metadata table {m.name!r}"
                                    for m in metadata
                                    for c in m.columns
                                     if c in output_source_columns]:
        raise AugurError(dedent(f"""\
            Generated source column names may not conflict with any column
            names in metadata inputs.

            The given source column template ({args.source_columns!r}) with the
            given metadata table names would conflict with the following input
            {_n("column", "columns", len(conflicting_columns))}:

              {indented_list(conflicting_columns, '            ' + '  ')}

            Please adjust the source column template with --source-columns
            and/or adjust the metadata table names to avoid conflicts.
            """))


    try:
        # Read all metadata files into a SQLite db
        for m in metadata:
            # All other metadata reading in Augur (i.e. via the csv module)
            # uses Python's "universal newlines"¹ definition and accepts \n,
            # \r\n, and \r as newlines interchangably (even mixed within the
            # same file!).  We accomplish the same behaviour here with SQLite's
            # less flexible newline handling by relying on the universal
            # newline translation of `augur read-file`.
            #   -trs, 24 July 2024
            #
            # ¹ <https://docs.python.org/3/glossary.html#term-universal-newlines>
            newline = os.linesep

            print_info(f"Reading {m.name!r} metadata from {m.path!r}…")
            sqlite3(db_path,
                f'.mode csv',
                f'.separator {sqlite_quote_dot(m.delimiter)} {sqlite_quote_dot(newline)}',
                f'.import {sqlite_quote_dot(f"|{augur} read-file {shquote(m.path)}")} {sqlite_quote_dot(m.table_name)}',

                f'create unique index {sqlite_quote_id(f"{m.table_name}_id")} on {sqlite_quote_id(m.table_name)}({sqlite_quote_id(m.id_column)});',

                # <https://sqlite.org/pragma.html#pragma_optimize>
                f'pragma optimize;')

            # We're going to use Metadata.columns to generate the select
            # statement, so ensure it matches what SQLite's .import created.
            assert m.columns == (table_columns := sqlite3_table_columns(db_path, m.table_name)), \
                f"{m.columns!r} == {table_columns!r}"

            # Track which columns appear in which metadata inputs, preserving
            # the order of both.
            for column in m.columns:
                # Match different id column names in different metadata files
                # since they're logically equivalent.  Any non-id columns that
                # match the output_id_column (i.e. first table's id column) and
                # would thus overwrite it with this logic are already a fatal
                # error above.
                output_column = output_id_column if column == m.id_column else column

                output_columns.setdefault(output_column, [])
                output_columns[output_column] += [(m.table_name, column)]


        # Construct query to produce merged metadata.
        select_list = [
            # Output metadata columns coalesced across input metadata columns
            *(f"""coalesce({', '.join(f"nullif({x}, '')" for x in starmap(sqlite_quote_id, reversed(input_columns)))}, null) as {sqlite_quote_id(output_column)}"""
                for output_column, input_columns in output_columns.items()),

            # Source columns.  Select expressions generated here instead of
            # earlier to stay adjacent to the join conditions below, upon which
            # these rely.
            *(f"""{sqlite_quote_id(m.table_name, m.id_column)} is not null as {sqlite_quote_id(output_source_column(m.name))}"""
                for m in metadata if output_source_column)]

        from_list = [
            sqlite_quote_id(metadata[0].table_name),
            *(f"full outer join {sqlite_quote_id(m.table_name)} on {sqlite_quote_id(m.table_name, m.id_column)} in ({', '.join(sqlite_quote_id(m.table_name, m.id_column) for m in reversed(preceding))})"
                for m, preceding in [(m, metadata[:i]) for i, m in enumerate(metadata[1:], 1)])]

        # Take some small pains to make the query readable since it makes
        # debugging and development easier.  Note that backslashes aren't
        # allowed inside f-string expressions, hence the *newline* variable.
        newline = '\n'
        query = dedent(f"""\
            select
                {(',' + newline + '                ').join(select_list)}
            from
                {(newline + '                ').join(from_list)}
            ;
            """)


        # Write merged metadata as export from SQLite db.
        #
        # Assume TSV like nearly all other extant --output-metadata options.
        print_info(f"Merging metadata and writing to {args.output_metadata!r}…")
        print_debug(query)
        sqlite3(db_path,
            f'.mode csv',
            f'.separator "\\t" "\\n"',
            f'.headers on',
            f'.once {sqlite_quote_dot(f"|{augur} write-file {shquote(args.output_metadata)}")}',
            query)

    except SQLiteError as err:
        delete_db = False
        raise AugurError(str(err)) from err

    finally:
        if delete_db:
            os.unlink(db_path)
        else:
            print_info(f"WARNING: Skipped deletion of {db_path} due to error, but you may want to clean it up yourself (e.g. if it's large).")


def sqlite3(*args, **kwargs):
    """
    Internal helper for invoking ``sqlite3``, the SQLite CLI program.
    """
    sqlite3 = os.environ.get("SQLITE3", which("sqlite3"))

    if not sqlite3:
        raise AugurError(dedent(f"""\
            Unable to find the program `sqlite3`.  Is it installed?

            In order to use `augur merge`, the SQLite 3 CLI (version ≥3.39)
            must be installed separately.  It is typically provided by a
            Nextstrain runtime.
            """))

    argv = [sqlite3, "-init", os.devnull, "-batch", *args]

    print_debug(f"running {argv!r}")
    proc = subprocess.run(argv, encoding="utf-8", text=True, **kwargs)

    try:
        proc.check_returncode()
    except subprocess.CalledProcessError as err:
        raise SQLiteError(f"sqlite3 invocation failed") from err

    return proc


class SQLiteError(Exception):
    pass


def sqlite3_table_columns(db_path, table: str) -> Iterable[str]:
    return sqlite3(db_path, f"select name from pragma_table_info({sqlite_quote_string(table)})", capture_output=True).stdout.splitlines();


def sqlite_quote_id(*xs):
    """
    Quote a SQLite identifier.

    <https://sqlite.org/lang_keywords.html>

    >>> sqlite_quote_id('foo bar')
    '"foo bar"'
    >>> sqlite_quote_id('table name', 'column name')
    '"table name"."column name"'
    >>> sqlite_quote_id('weird"name')
    '"weird""name"'
    """
    return '.'.join('"' + x.replace('"', '""') + '"' for x in xs)


def sqlite_quote_dot(x):
    """
    Quote a SQLite CLI dot-command argument.

    <https://sqlite.org/cli.html#dot_command_arguments>
    """
    return '"' + x.replace('\\', '\\\\').replace('"', '\\"') + '"'


def sqlite_quote_string(x):
    """
    Quote a SQLite string (i.e. produce a string literal).

    <https://www.sqlite.org/lang_expr.html#literal_values_constants_>
    """
    return "'" + x.replace("'", "''") + "'"


def pairs(xs: Iterable[str]) -> Iterable[Tuple[str, str]]:
    """
    Split an iterable of ``k=v`` strings into an iterable of ``(k,v)`` tuples.

    >>> pairs(["abc=123", "eight nine ten=el em en"])
    [('abc', '123'), ('eight nine ten', 'el em en')]

    Strings missing a ``k`` and/or a ``v`` part get an empty string.

    >>> pairs(["v", "=v", "k=", "=", ""])
    [('', 'v'), ('', 'v'), ('k', ''), ('', ''), ('', '')]

    ``k`` ends at the first ``=``.

    >>> pairs(["abc=123=xyz", "=v=v"])
    [('abc', '123=xyz'), ('', 'v=v')]
    """
    return [tuple(x.split("=", 1)) if "=" in x else ("", x) for x in xs] # type: ignore


def count_unique(xs: Iterable[T]) -> Iterable[Tuple[T, int]]:
    # Using reduce() with a dict because it preserves input order, unlike
    # itertools.groupby(), which requires a sort.  Preserving order is a nice
    # property for the user since we generate an error message with this.
    #   -trs, 24 July 2024
    yield from reduce(lambda counts, x: {**counts, x: counts.get(x, 0) + 1}, xs, counts := {}).items() # type: ignore


def indented_list(xs, prefix):
    return f"\n{prefix}".join(xs)


def shquote_humanized(x):
    r"""
    shquote for humans.

    Use C-style escapes supported by shells (specifically, Bash) for characters
    that humans would typically use C-style escapes for instead of quoted
    literals.

    <https://www.gnu.org/software/bash/manual/bash.html#ANSI_002dC-Quoting>

    >>> shquote_humanized("abc")
    'abc'

    >>> shquote_humanized("\t")
    "$'\\t'"

    >>> shquote_humanized("abc def")
    "'abc def'"

    >>> shquote_humanized("abc\tdef")
    "abc$'\\t'def"
    """
    escapes = {
        '\a': r'\a',
        '\b': r'\b',
        '\f': r'\f',
        '\n': r'\n',
        '\r': r'\r',
        '\t': r'\t',
        '\v': r'\v',
    }

    def quote(s):
        if s in escapes:
            return f"$'{escapes[s]}'"
        else:
            # split leaves leading and trailing empty strings when its input is
            # entirely (captured) separator.  Avoid quoting every empty string
            # *part* here…
            return shquote(s) if s else ''

    parts = re.split('([' + ''.join(escapes.values()) + '])', x)
    quoted = ''.join(map(quote, parts))

    # …and instead quote a final empty string down here if we're still empty
    # after joining all our parts together.
    return quoted if quoted else shquote('')



import argparse
import datetime
from textwrap import dedent
import isodate
import pandas as pd
import re
import treetime.utils
from augur.errors import AugurError
from .errors import InvalidDate

from .ambiguous_date import AmbiguousDate

SUPPORTED_DATE_HELP_TEXT = dedent("""\
    1. an Augur-style numeric date with the year as the integer part (e.g. 2020.42) or
    2. a date in ISO 8601 date format (i.e. YYYY-MM-DD) (e.g. '2020-06-04') or
    3. a backwards-looking relative date in ISO 8601 duration format with optional P prefix (e.g. '1W', 'P1W')
""")

def numeric_date(date):
    """
    Converts the given *date* string to a :py:class:`float`.

    *date* may be given as:
    1. A string or float (number) with year as the integer part
    2. A string in the YYYY-MM-DD (ISO 8601) syntax
    3. A string representing a relative date (duration before datetime.date.today())

    Examples
    --------
    >>> numeric_date("2020.42")
    2020.42
    >>> numeric_date("2020-06-04")
    2020.42486...
    >>> import datetime, isodate, treetime
    >>> numeric_date("1W") == treetime.utils.numeric_date(datetime.date.today() - isodate.parse_duration("P1W"))
    True
    """
    # date is a datetime.date
    if isinstance(date, datetime.date):
        return treetime.utils.numeric_date(date)

    # date is numeric
    try:
        return float(date)
    except ValueError:
        pass

    # date is in YYYY-MM-DD form
    try:
        return treetime.utils.numeric_date(datetime.date(*map(int, date.split("-", 2))))
    except ValueError:
        pass

    # date is a duration treated as a backwards-looking relative date
    try:
        # make a copy of date for this block
        duration_str = str(date)
        if duration_str.startswith('P'):
            duration_str = duration_str
        else:
            duration_str = 'P'+duration_str
        return treetime.utils.numeric_date(datetime.date.today() - isodate.parse_duration(duration_str))
    except (ValueError, isodate.ISO8601Error):
        pass

    raise InvalidDate(date, f"""Ensure it is in one of the supported formats:\n{SUPPORTED_DATE_HELP_TEXT}""")

def numeric_date_type(date):
    """Wraps numeric_date() for argparse usage.

    This raises an ArgumentTypeError from InvalidDateFormat exceptions, otherwise the custom exception message won't be shown in console output due to:
    https://github.com/python/cpython/blob/5c4d1f6e0e192653560ae2941a6677fbf4fbd1f2/Lib/argparse.py#L2503-L2513
    """
    try:
        return numeric_date(date)
    except InvalidDate as error:
        raise argparse.ArgumentTypeError(str(error)) from error

def is_date_ambiguous(date, ambiguous_by):
    """
    Returns whether a given date string in the format of YYYY-MM-DD is ambiguous by a given part of the date (e.g., day, month, year, or any parts).

    Parameters
    ----------
    date : str
        Date string in the format of YYYY-MM-DD
    ambiguous_by : str
        Field of the date string to test for ambiguity ("day", "month", "year", "any")
    """
    date_components = date.split('-', 2)

    if len(date_components) == 3:
        year, month, day = date_components
    elif len(date_components) == 2:
        year, month = date_components
        day = "XX"
    else:
        year = date_components[0] if date_components[0] else 'X'
        month = "XX"
        day = "XX"

    # Determine ambiguity hierarchically such that, for example, an ambiguous
    # month implicates an ambiguous day even when day information is available.
    return any((
        "X" in year,
        "X" in month and ambiguous_by in ("any", "month", "day"),
        "X" in day and ambiguous_by in ("any", "day")
    ))

def get_numerical_date_from_value(value, fmt=None, min_max_year=None):
    value = str(value)
    if re.match(r'^-*\d+\.\d+$', value):
        # numeric date which can be negative
        return float(value)
    if value.isnumeric():
        # year-only date is ambiguous
        value = fmt.replace('%Y', value).replace('%m', 'XX').replace('%d', 'XX')
    if 'XX' in value:
        try:
            ambig_date = AmbiguousDate(value, fmt=fmt).range(min_max_year=min_max_year)
        except InvalidDate as error:
            raise AugurError(str(error)) from error
        return [treetime.utils.numeric_date(d) for d in ambig_date]
    try:
        return treetime.utils.numeric_date(datetime.datetime.strptime(value, fmt))
    except:
        return None

def get_numerical_dates(metadata:pd.DataFrame, name_col = None, date_col='date', fmt=None, min_max_year=None):
    if not isinstance(metadata, pd.DataFrame):
        raise AugurError("Metadata should be a pandas.DataFrame.")
    if fmt:
        strains = metadata.index.values
        dates = metadata[date_col].apply(
            lambda date: get_numerical_date_from_value(
                date,
                fmt,
                min_max_year
            )
        ).values
    else:
        strains = metadata.index.values
        dates = metadata[date_col].astype(float)
    return dict(zip(strains, dates))

def get_year_month(year, month):
    return f"{year}-{str(month).zfill(2)}"

def get_year_week(year, month, day):
    year, week = datetime.date(year, month, day).isocalendar()[:2]
    return f"{year}-{str(week).zfill(2)}"



"""
Functions that are from earlier versions of `augur validate` and haven't yet been
moved into `validate.py`. There's a lot of duplication here which can be abstracted
and refactored over time.
"""

import sys
from collections import defaultdict

def ensure_no_duplicate_names(root, ValidateError):
    """
    Check that all node names are identical, which is required for auspice (v2) JSONs.
    """
    names = set()
    def recurse(node):
        if node["name"] in names:
            raise ValidateError(f"Node {node['name']} appears multiple times in the tree.")
        names.add(node["name"])
        if "children" in node:
            [recurse(child) for child in node["children"]]
    recurse(root)


def collectTreeAttrsV2(root, warn):
    """
    Collect all keys specified on `node["node_attrs"]` throughout the tree
    and the values associated with them. Note that this will only look at
    attributes which are themselves objects with a `value` property.
    I.e. a node attribute `node["node_attrs"]["div"] -> numeric` will not
    be collected.
    Returns a tuple.
    return[0]: dict of `node_attr_property` -> x, where x is a dict with
    keys `count` -> INT, `values` -> SET, `onAllNodes` -> BOOL.
    return[1]: INT of number of terminal nodes in tree
    """
    seen = defaultdict(lambda: {"count": 0, "values": set(), "onAllNodes": False})
    num_nodes, num_terminal = (0, 0)
    def recurse(node):
        nonlocal num_nodes, num_terminal
        num_nodes += 1
        for prop, info in node.get("node_attrs", {}).items():
            if not isinstance(info, dict) or "value" not in info:
                continue
            seen[prop]["count"] += 1
            seen[prop]["values"].add(info["value"])
        if "children" in node:
            [recurse(child) for child in node["children"]]
        else:
            num_terminal += 1
    recurse(root)

    for data in seen.values():
        if data["count"] == num_nodes:
            data["onAllNodes"] = True

    return(seen, num_terminal)


def collectMutationGenes(root):
    """
    Returns a set of all genes specified in the tree in the "aa_muts" objects
    """
    genes = set()
    def recurse(node):
        mutations = node.get("branch_attrs", {}).get("mutations", False)
        if mutations:
            genes.update(mutations.keys())
        if "children" in node:
            [recurse(child) for child in node["children"]]
    recurse(root)
    genes -= {"nuc"}
    return genes

def collectBranchLabels(root):
    labels = set()
    def recurse(node):
        labels.update(node.get("branch_attrs", {}).get("labels", {}).keys())
        if "children" in node:
            [recurse(child) for child in node["children"]]
    recurse(root)
    return labels

def verifyMainJSONIsInternallyConsistent(data, ValidateError):
    """
    Check possible sources of conflict within the main (unified) JSON
    This function is only used for schema v2.0
    In an ideal world, the JSON schema would be able to validate this,
    however this function performs tests which are not possible to
    define in the schema (as of JSON schema v6).
    """
    warnings = False
    def warn(msg):
        nonlocal warnings
        warnings = True
        print("\tWARNING: ", msg, file=sys.stderr)

    print("Validating that the JSON is internally consistent...")

    ensure_no_duplicate_names(data["tree"], ValidateError)

    if "entropy" in data["meta"]["panels"] and "genome_annotations" not in data["meta"]:
        warn("The entropy panel has been specified but annotations don't exist.")

    tree_traits, _ = collectTreeAttrsV2(data["tree"], warn)

    if "geo_resolutions" in data["meta"]:
        for geo_res in data["meta"]["geo_resolutions"]:
            geo_name = geo_res["key"]
            deme_to_lat_longs = geo_res["demes"]
            if geo_name not in tree_traits:
                warn("The geographic resolution \"{}\" does not appear on any tree nodes.".format(geo_name))
                continue
            # pass 1: check the demes in "geo_resolutions" are found on the tree
            for geoValue in deme_to_lat_longs.keys():
                if geoValue not in tree_traits[geo_name]["values"]:
                    warn("\"{}\", a value of the geographic resolution \"{}\", does not appear on any tree nodes.".format(geoValue, geo_name))
            # pass 1: check the demes across the tree are represented in "geo_resolutions"
            for geoValue in tree_traits[geo_name]["values"]:
                if geoValue not in deme_to_lat_longs:
                    warn("\"{}\", a value of the geographic resolution \"{}\", appears in the tree but not in the metadata."
                        "\n\t\tThis will cause transmissions & demes involving this location not to be displayed in Auspice".format(geoValue, geo_name))
    else:
        if "map" in data["meta"]["panels"]:
            warn("Map panel was requested but no geographic_info was provided")

    if "colorings" in data["meta"]:
        for coloring in data["meta"]["colorings"]:
            colorBy = coloring["key"]
            if colorBy == "gt":
                continue
            if colorBy not in tree_traits:
                warn("The coloring \"{}\" does not appear as an attr on any tree nodes.".format(colorBy))
            if "scale" in coloring:
                scale = coloring["scale"]
                if isinstance(scale, list):
                    for value, hex in scale:
                        if value not in tree_traits[colorBy]["values"]:
                            warn("Color option \"{}\" specifies a hex code for \"{}\" but this isn't ever seen on the tree nodes.".format(colorBy, value))
                elif isinstance(scale, str):
                    raise ValidateError("String colour scales are not yet implemented")
                else:
                    raise ValidateError("Invalid color scale (for trait \"{}\")".format(colorBy))
            if "domain" in coloring:
                domain = coloring["domain"]
                if coloring["type"] in ["ordinal", "categorical"]:
                    inMetaNotInTree = [val for val in domain if val not in tree_traits[colorBy]["values"]]
                    if len(inMetaNotInTree):
                        warn("Domain for {} defined the following values which are not present on the tree: {}".format(colorBy, inMetaNotInTree.join(", ")))
                    inTreeNotInMeta = [val for val in tree_traits[colorBy]["values"] if val not in domain]
                    if len(inTreeNotInMeta):
                        warn("Tree defined values for {} which were not in the domain: {}".format(colorBy, inTreeNotInMeta.join(", ")))
                elif coloring["type"] == "boolean":
                    raise ValidateError("Cannot povide a domain for a boolean coloring ({})".format(colorBy))
    else:
        warn("No colourings were provided")

    if "filters" in data["meta"]:
        for filter in data["meta"]["filters"]:
            if filter not in tree_traits:
                warn("The filter \"{}\" does not appear as a property on any tree nodes.".format(filter))

    genes_with_mutations = collectMutationGenes(data['tree'])
    if len(genes_with_mutations):
        if "genome_annotations" not in data["meta"]:
            warn("The tree defined mutations on genes {}, but annotations aren't defined in the meta JSON.".format(", ".join(genes_with_mutations)))
        else:
            for gene in genes_with_mutations:
                if gene not in data["meta"]["genome_annotations"]:
                    warn("The tree defined mutations on gene {} which doesn't appear in the metadata annotations object.".format(gene))

    default_branch_label = data.get("meta").get("display_defaults", {}).get("branch_label")
    if default_branch_label and default_branch_label.lower() != "none":
        labels = collectBranchLabels(data['tree'])
        if not default_branch_label in labels:
            warn("Default label to display \"{}\" isn't found anywhere on the tree!".format(default_branch_label))

    return not warnings


def collectTreeAttrsV1(root):
    """
    Collect all keys specified on node->attr (or node->traits) throughout the tree
    If the values of these keys are strings, then also collect the values
    """
    seen = defaultdict(lambda: {"count": 0, "values": set(), "onAllNodes": False})
    num_nodes, num_terminal = (0, 0)
    def recurse(node):
        nonlocal num_nodes, num_terminal
        num_nodes += 1
        traits = node["attr"].keys()
        for property in traits:
            seen[property]["count"] += 1
            value = node["attr"][property]
            if isinstance(value, str):
                seen[property]["values"].add(value)

        if "children" in node:
            [recurse(child) for child in node["children"]]
        else:
            num_terminal += 1
    recurse(root)

    for data in seen.values():
        if data["count"] == num_nodes:
            data["onAllNodes"] = True

    return(seen, num_terminal)


def collectAAMutationGenesV1(root):
    """
    Returns a set of all genes specified in the tree in the "aa_muts" objects
    """
    genes = set()
    def recurse(node):
        if "aa_muts" in node and len(node["aa_muts"]):
            genes.update(node["aa_muts"].keys())
        if "children" in node:
            [recurse(child) for child in node["children"]]
    recurse(root)
    return genes


def verifyMetaAndOrTreeJSONsAreInternallyConsistent(meta_json, tree_json, ValidateError):
    """
    Check all possible sources of conflict internally & between the metadata & tree JSONs
    This is only that which cannot be checked by the schemas
    """
    warnings = False
    def warn(msg):
        nonlocal warnings
        warnings = True
        print("\tWARNING: ", msg, file=sys.stderr)


    print("Validating that meta + tree JSONs are internally consistent...")
    mj = meta_json

    if "panels" in mj and "entropy" in mj["panels"] and "annotations" not in mj:
        warn("\tERROR: The entropy panel has been specified but annotations don't exist.")

    tj = tree_json
    treeAttrs, num_terminal_nodes = collectTreeAttrsV1(tj)

    if "geo" in mj:
        for geoName in mj["geo"]:
            if geoName not in treeAttrs:
                warn("The geographic resolution \"{}\" does not appear as an attr on any tree nodes.".format(geoName))
            else:
                for geoValue in mj["geo"][geoName]:
                    if geoValue not in treeAttrs[geoName]["values"]:
                        warn("\"{}\", a value of the geographic resolution \"{}\", does not appear as a value of attr->{} on any tree nodes.".format(geoValue, geoName, geoName))
                for geoValue in treeAttrs[geoName]["values"]:
                    if geoValue not in mj["geo"][geoName]:
                        warn("\"{}\", a value of the geographic resolution \"{}\", appears in the tree but not in the metadata.".format(geoValue, geoName))
                        warn("\tThis will cause transmissions & demes involving this location not to be displayed in Auspice")


    if "color_options" in mj:
        for colorBy in [x for x in mj["color_options"] if x != "gt"]:
            if colorBy not in treeAttrs:
                warn("The color_option \"{}\" does not appear as an attr on any tree nodes.".format(colorBy))
            elif "color_map" in mj["color_options"][colorBy]:
                # if there's a color_map, then check that there are no values which aren't seen on the tree
                for (value, hex) in mj["color_options"][colorBy]["color_map"]:
                    if value not in treeAttrs[colorBy]["values"]:
                        warn("Color option \"{}\" specifies a hex code for \"{}\" but this isn't ever seen on the tree nodes.".format(colorBy, value))
                # inversely, check for values on the tree not defined in the color_map
                for value in treeAttrs[colorBy]["values"]:
                    color_map_values = [x[0] for x in mj["color_options"][colorBy]["color_map"]]
                    if value not in color_map_values:
                        warn("Color option \"{}\", which contains a color_map, is missing \"{}\"".format(colorBy, value))



    if "filters" in mj:
        for filter in mj["filters"]:
            if filter not in treeAttrs:
                warn("The filter \"{}\" does not appear on any tree nodes.".format(filter))


    if "virus_count" in mj and mj["virus_count"] != num_terminal_nodes:
        raise ValidateError("Meta JSON virus_count ({}) differs from the number of nodes in the tree ({})".format(mj["virus_count"], num_terminal_nodes))

    genes_with_aa_muts = collectAAMutationGenesV1(tj)
    if len(genes_with_aa_muts):
        if "annotations" not in mj:
            warn("The tree defined AA mutations on genes {}, but annotations aren't defined in the meta JSON.".format(", ".join(genes_with_aa_muts)))
        else:
            for gene in genes_with_aa_muts:
                if gene not in mj["annotations"]:
                    warn("The tree defined AA mutations on gene {} which doesn't appear in the metadata annotations object.".format(gene))

    return not warnings



import json

from augur.__version__ import __version__
from augur.__version__ import is_augur_version_compatible
from augur.errors import AugurError
from augur.io.file import open_file
from augur.io.print import print_err
from augur.types import ValidationMode
from augur.validate import validate_json, ValidateError, load_json_schema


FILTERED_ATTRS = ["generated_by"]


class NodeDataFile:
    def __init__(self, fname, validation_mode=ValidationMode.ERROR):
        self.fname = fname
        self.validation_mode = validation_mode

        with open_file(fname) as jfile:
            self.attrs = json.load(jfile)

        self.validate()

    @property
    def annotations(self):
        return self.attrs.get("annotations")

    @property
    def nodes(self):
        return self.attrs.get("nodes", {})

    @property
    def branches(self):
        # these are optional, so we provide an empty dict as a default
        return self.attrs.get("branches", {})

    @property
    def generated_by(self):
        return self.attrs.get("generated_by")

    @property
    def is_generated_by_incompatible_augur(self):
        if not isinstance(self.generated_by, dict):
            # If it's not a dict created by augur, we can't reliably classify it as incompatible
            return False

        generated_by_augur = self.generated_by.get("program") == "augur"
        compatible_version = is_augur_version_compatible(
            self.generated_by.get("version")
        )

        return generated_by_augur and not compatible_version

    def items(self):
        filtered_attrs = {
            key: value for key, value in self.attrs.items() if key not in FILTERED_ATTRS
        }

        return filtered_attrs.items()

    def validate(self):
        if self.validation_mode is ValidationMode.SKIP:
            return # don't perform validation (i.e. skip it)

        try:
            if self.annotations:
                try:
                    # validate_json will print any errors to STDERR
                    validate_json(
                        self.annotations,
                        load_json_schema("schema-annotations.json"),
                        self.fname,
                    )
                except ValidateError as err:
                    raise ValidateError(
                        f"{self.fname} contains an `annotations` attribute of an invalid JSON format. Was it "
                        "produced by different version of augur the one you are currently using "
                        f" ({__version__})? Please check the program that produced that JSON file."
                    ) from err

            if not isinstance(self.nodes, dict):
                raise ValidateError(
                    f"`nodes` value in {self.fname} is not a dictionary. Please check the formatting of this JSON!"
                )

            if not isinstance(self.branches, dict):
                raise ValidateError(
                    f"`branches` value in {self.fname} is not a dictionary. Please check the formatting of this JSON!"
                )

            if not self.nodes and not self.branches:
                print_err(
                    f"WARNING: {self.fname} has empty or nonexistent `nodes` and `branches`. Please check the formatting of this JSON!"
                )

            if self.is_generated_by_incompatible_augur:
                msg = (
                    f"Augur version incompatibility detected: the JSON {self.fname} was generated by "
                    f"{self.generated_by}, which is incompatible with the current augur version "
                    f"({__version__}). We suggest you rerun the pipeline using the current version of "
                    "augur."
                )
                raise ValidateError(msg)
        except ValidateError as e:
            if self.validation_mode is ValidationMode.WARN:
                # string-representation of error mirrors how augur ultimately handles AugurError
                print_err(f"WARNING: {e}")
            else:
                # Re-raising within an except block will use implicit chaining (via __context__)
                # however since we ultimately catch AugurError and use a string representation
                # this isn't particularly helpful. One day we may add a dev-flag to allow printing
                # the traceback or similar
                raise AugurError(*e.args)


class NodeDataObject(NodeDataFile):
    """
    NodeDataObject is identical to NodeDataFile except it takes a node-data dict
    rather than loading the node data from a file
    """
    def __init__(self, node_data_json, fname, validation_mode=ValidationMode.ERROR):
        self.fname = fname
        self.validation_mode = validation_mode
        self.attrs = node_data_json
        self.validate()



from augur.errors import AugurError
from augur.filter.weights_file import get_weighted_columns
from augur.io.vcf import is_vcf as filename_is_vcf


SEQUENCE_ONLY_FILTERS = (
    "min_length",
    "max_length",
    "non_nucleotide",
)


def validate_arguments(args):
    """Validate arguments.

    Parameters
    ----------
    args : argparse.Namespace
        Parsed arguments from argparse
    """
    # Don't allow sequence output when no sequence input is provided.
    if args.output and not args.sequences:
        raise AugurError("You need to provide sequences to output sequences.")

    # Confirm that at least one output was requested.
    if not any((args.output, args.output_metadata, args.output_strains)):
        raise AugurError("You need to select at least one output.")

    # Don't allow filtering on sequence-based information, if no sequences or
    # sequence index is provided.
    if not args.sequences and not args.sequence_index and any(getattr(args, arg) for arg in SEQUENCE_ONLY_FILTERS):
        raise AugurError("You need to provide a sequence index or sequences to filter on sequence-specific information.")

    # Set flags if VCF
    is_vcf = filename_is_vcf(args.sequences)

    # Confirm that vcftools is installed.
    if is_vcf:
        from shutil import which
        if which("vcftools") is None:
            raise AugurError("'vcftools' is not installed! This is required for VCF data. "
                  "Please see the augur install instructions to install it.")

    # If user requested grouping, confirm that other required inputs are provided, too.
    if args.group_by and not any((args.sequences_per_group, args.subsample_max_sequences)):
        raise AugurError("You must specify a number of sequences per group or maximum sequences to subsample.")
    
    # Weighted columns must be specified explicitly.
    if args.group_by_weights:
        weighted_columns = get_weighted_columns(args.group_by_weights)
        if (not set(weighted_columns) <= set(args.group_by)):
            raise AugurError("Columns in --group-by-weights must be a subset of columns provided in --group-by.")

    # --output-group-by-sizes is only available for --group-by-weights.
    if args.output_group_by_sizes and not args.group_by_weights:
        raise AugurError(
            "--output-group-by-sizes is only available for --group-by-weights. "
            "It may be added to other sampling methods in the future - see <https://github.com/nextstrain/augur/issues/1590>"
        )

    # --group-by-weights cannot be used with --no-probabilistic-sampling.
    if args.group_by_weights and not args.probabilistic_sampling:
        raise AugurError(
            "--group-by-weights cannot be used with --no-probabilistic-sampling."
        )



"""
Parses GenBank's 'location' field of the NDJSON record to 3 separate
fields: 'country', 'division', and 'location'.

Checks that a record is from GenBank by verifying that the 'database'
field has a value of "GenBank" or "RefSeq".
"""

import argparse
from typing import Generator, List
from augur.io.print import print_err
from augur.utils import first_line


def parse_location(
    record: dict,
    location_field_name: str,
) -> dict:
    # Expected pattern for the location field is
    # "<country_value>[:<region>][, <locality>]"
    #
    # See GenBank docs for their "country" field:
    # https://www.ncbi.nlm.nih.gov/genbank/collab/country/
    location_field = record.get(location_field_name, None)
    if location_field is None:
        print_err(
            f"`parse-genbank-location` requires a `{location_field_name}` field; this record does not have one.",
        )
        # bail early because we're not gonna make any changes
        return record

    geographic_data = location_field.split(":")

    country = geographic_data[0]
    division = ""
    location = ""

    if len(geographic_data) == 2:
        division, _, location = geographic_data[1].partition(",")

    record["country"] = country.strip()
    record["division"] = division.strip()
    record["location"] = location.strip()

    return record


def register_parser(
    parent_subparsers: argparse._SubParsersAction,
) -> argparse._SubParsersAction:
    parser = parent_subparsers.add_parser(
        "parse-genbank-location",
        parents=[parent_subparsers.shared_parser],  # type: ignore
        help=first_line(__doc__),
    )

    parser.add_argument(
        "--location-field",
        default="geo_loc_name",
        help="The field containing the location, "
        + "in the format `<geo_loc_name>[:<region>][, <locality>]`",
    )

    return parser


def run(
    args: argparse.Namespace,
    records: List[dict],
) -> Generator[dict, None, None]:
    for record in records:
        parse_location(
            record,
            args.location_field,
        )
        yield record



from augur.ancestral import run_ancestral
from io import StringIO
from Bio import Phylo
import pytest

def gather_mutations_at_pos(pos, nuc_result):
    """pos is 1-based"""
    muts = []
    for v in nuc_result['mutations']['nodes'].values():
        for mut in v['muts']:
            if int(mut[1:-1])==pos:
                muts.append(mut)
    return muts

def gather_bases_at_pos(pos, nuc_result):
    """pos is 1-based"""
    bases = set()
    for v in nuc_result['mutations']['nodes'].values():
        bases.add(v['sequence'][pos-1])
    return bases

class TestRootNodeMutationAssignment:
    @pytest.fixture
    def tree(self):
        t = "(sample_C:0.02,(sample_B:0.02,sample_A:0.02)node_AB:0.06)node_root:0.02;"
        T = Phylo.read(StringIO(t), 'newick')
        return T

    @pytest.fixture()
    def ref(self):
        return "AAAAAA"
        #       123456

    @pytest.fixture()
    def default_augur_args(self):
        return {
            'fill_overhangs': True, # augur default
            'marginal': 'joint', # augur default
            'alphabet': 'nuc', # augur default
        }


    @pytest.fixture()
    def fasta_aln(self):
        from Bio.Seq import Seq
        from Bio.SeqRecord import SeqRecord
        from Bio.Align import MultipleSeqAlignment
        aln = MultipleSeqAlignment(
            [
                SeqRecord(Seq("AANATA"), id="sample_A"),
                SeqRecord(Seq("AANATG"), id="sample_B"),
                SeqRecord(Seq("AANCCA"), id="sample_C"),
            ]
        )
        return aln

    @pytest.fixture()
    def vcf_aln(self):
        def to_zero(x):
            return int(x)-1
        return {
            'sample_A': {to_zero(3): 'N', to_zero(5): 'T'},
            'sample_B': {to_zero(3): 'N', to_zero(5): 'T', to_zero(6): 'G'},
            'sample_C': {to_zero(3): 'N', to_zero(4): 'C', to_zero(5): 'C'}
        }

    def test_fasta_with_ambiguity_inferred(self, tree, fasta_aln, ref, default_augur_args):
        is_vcf = False
        nuc_result = run_ancestral(T=tree, aln=fasta_aln, reference_sequence=ref, is_vcf=is_vcf, full_sequences=not is_vcf, infer_ambiguous=True, **default_augur_args)
        assert(nuc_result['root_seq'] == ref)
        assert(gather_bases_at_pos(3, nuc_result)==set(['A']))
        assert(gather_mutations_at_pos(3, nuc_result)==[])

    def test_fasta_with_ambiguity_not_inferred(self, tree, fasta_aln, ref, default_augur_args):
        is_vcf = False
        nuc_result = run_ancestral(T=tree, aln=fasta_aln, reference_sequence=ref, is_vcf=is_vcf, full_sequences=not is_vcf, infer_ambiguous=False, **default_augur_args)
        assert(nuc_result['root_seq'] == ref)
        assert(gather_bases_at_pos(3, nuc_result)==set(['N']))
        # Expected behaviour here is that every tip is "N", and this is also
        # inferred over all the internal nodes (so they're all "N") and thus we have
        # a single mutation "C3N" on the root node. This is not what happens - the
        # full sequences on each node are like this, but the reported mutations are
        # incorrect - we are describing a "C3N" on each terminal node. I'm removing
        # this test now as (a) it's not newly introduced and (b) I don't have the
        # time to fix it right now.
        # Note that the vcf version of this has the same bug.
        # assert(gather_mutations_at_pos(3, nuc_result)==['A3N'])

    def test_vcf_with_ambiguity_inferred(self, tree, vcf_aln, ref, default_augur_args):
        is_vcf = True
        nuc_result = run_ancestral(T=tree, aln=vcf_aln, reference_sequence=ref, is_vcf=is_vcf, full_sequences=not is_vcf, infer_ambiguous=True, **default_augur_args)
        assert(nuc_result['root_seq'] == ref)
        assert(gather_mutations_at_pos(3, nuc_result)==[])



import Bio.Phylo
import sys

from augur.types import ValidationMode
from augur.util_support.node_data import DuplicatedNonDictAttributeError
from augur.util_support.node_data import NodeData
from augur.util_support.node_data_file import NodeDataFile


class NodeDataReader:
    """
    Parses one or more node data files and combines them into a single data structure.

    The format of Node data files is described in `docs/usage/json_format.md`. The `generated_by` attribute is
    not included in the resulting data structure.

    If a tree file is specified, it is used to verify the node names.

    If validation_mode is set to :py:attr:`augur.types.ValidationMode.SKIP` no validation is performed.
    """

    def __init__(self, filenames, tree_file=None, validation_mode=ValidationMode.ERROR):
        if not isinstance(filenames, list):
            filenames = [filenames]
        self.filenames = filenames
        self.tree_file = tree_file
        self.validation_mode = validation_mode

    def read(self):
        node_data = self.build_node_data()

        self.check_against_tree_file(node_data)

        return node_data

    def build_node_data(self):
        node_data = NodeData()

        for node_data_file in self.node_data_files:
            try:
                node_data.update(node_data_file)
            except DuplicatedNonDictAttributeError as e:
                print(
                    f"Error parsing {node_data_file.fname}: Mixed dict and non-dict values found for "
                    f"attribute `{e.key}`. Augur requires attributes of the same name to be uniformly dicts "
                    "or non-dicts.",
                    file=sys.stderr,
                )
                sys.exit(2)

        return node_data.attrs

    @property
    def node_data_files(self):
        return (NodeDataFile(fname, validation_mode = self.validation_mode) for fname in self.filenames)

    def check_against_tree_file(self, node_data):
        if not self.tree_file:
            return

        if set(node_data["nodes"].keys()) != self.node_names_from_tree_file:
            print(
                f"Names of nodes (including internal nodes) of tree {self.tree_file} don't match node names "
                "in the node data files.",
                file=sys.stderr,
            )
            sys.exit(2)

    @property
    def node_names_from_tree_file(self):
        try:
            tree = Bio.Phylo.read(self.tree_file, "newick")
        except Exception as e:
            print(
                f"Failed to read tree from file {self.tree_file}: {e}", file=sys.stderr
            )
            sys.exit(2)

        return set([clade.name for clade in tree.find_clades()])
